/* -----------------------------------------------------------------------------
*                          SYSTEMSOFT MODULE HEADER
*  -----------------------------------------------------------------------------
*                                DELIVERABLE
*
*  CLIENT                : Comtrol Europe  Limited
*  PROJECT               : CMT/25 : Rocketport 550 drivers for Linux
*
*  CLASSIFICATION        : Commercial In Confidence
*
*  FILE NAME             : $Workfile: RCKT550.C $
*  MODULE ISSUE          : $Revision: 4 $
*
*  LANGUAGE              : C
*
*  HOST MACHINE          : Pentium Class running Linux 7.2 or later.
*
*  TARGET MACHINE        : Pentium Class running Linux 7.2 or later.
*
*  -----------------------------------------------------------------------------
*
*  MODULE FUNCTION       :
*
*  REQUIREMENT SPEC(S)   : Not Applicable
*
*  DESIGN SPEC(S)        : Not Applicable
*
*  CONFIGURATION TOOL    : Microsoft Visual Sourcesafe 5.0
*
*  COMMENTS              :
*
*  -----------------------------------------------------------------------------
*  AUTHOR                : $Author: Stan $
*  CREATION DATE         : 11/7/2000
*  -----------------------------------------------------------------------------
*
*                         MODIFICATION & REVIEW RECORD
* $History: RCKT550.C $
 * 
 * *****************  Version 4  *****************
 * User: Stan         Date: 16/10/03   Time: 16:10
 * Updated in $/Comtrol/RocketPort550/Linux
 * Fix interrupt sharing allocation/free issues. Remove timers as all PCI
 * cards are interrupt-driven. Remove non-Rocketport hardware
 * dependencies. Enforce IRQ sharing.
 * 
 * *****************  Version 2  *****************
 * User: Stan         Date: 28/05/03   Time: 9:36
 * Updated in $/Comtrol/RocketPort550/Linux
 * RocketPort 550 Universal 
 * 
 * *****************  Version 1  *****************
 * User: Stan         Date: 25/04/03   Time: 10:37
 * Created in $/Comtrol/RocketPort550/Linux
 * 
 * *****************  Version 2  *****************
 * User: Stan         Date: 15/04/03   Time: 11:14
 * Updated in $/Comtrol/RocketPort 550 Linux
 * 
 * *****************  Version 1  *****************
 * User: Stan         Date: 21/02/01   Time: 10:46
 * Created in $/Comtrol/RocketPort 550 Linux
 * First Issue
*
*  -----------------------------------------------------------------------------
*/
/*
 *  linux/drivers/char/rckt550.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92.  Now
 *  much more extensible to support other rckt550 cards based on the
 *  16450/16550A UART's.  Added support for the AST FourPort and the
 *  Accent Async board.
 *
 *  set_serial_info fixed to set the flags, custom divisor, and uart
 * 	type fields.  Fix suggested by Michael K. Johnson 12/12/92.
 *
 *  11/95: TIOCMIWAIT, TIOCGICOUNT by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  03/96: Modularised by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  rckt550_set_termios fixed to look also for changes of the input
 *      flags INPCK, BRKINT, PARMRK, IGNPAR and IGNBRK.
 *                                            Bernd Anhäupl 05/17/96.
 *
 *  1/97:  Extended dumb rckt550 ports are a config option now.
 *         Saves 4k.   Michael A. Griffith <grif@acm.org>
 *
 *  8/97: Fix bug in rckt550_set_termios with RTS
 *        Stanislav V. Voronyi <stas@uanet.kharkov.ua>
 *
 *  3/98: Change the IRQ detection, use of probe_irq_o*(),
 *	  supress TIOCSERGWILD and TIOCSERSWILD
 *	  Etienne Lorrain <etienne.lorrain@ibm.net>
 *
 *  4/98: Added changes to support the ARM architecture proposed by
 * 	  Russell King
 *
 ********************************************************************************
	Comtrol Europe RocketPort 550 
	Modification History

   V1.00  01/01   SJB  SystemSoft Original driver release. 2.2.x kernels
   V1.01  08/01   LZ   EGO Update to support 2.4.x kernels
   V1.02  08/02   RLH  Fix build problem with 2.4.18 kernel
   V1.03  01/03   RLH  1. Documentation (readme) change
                       2. Add conditional to use slab.h in place of malloc.h 
                             (malloc deprecated in RH7.3 and removed from RH8.0)
                       3. Update Makefile to use new gcc flags

   V2.00  08/03   SJB  Add support for RocketPort Universal boards

   V2.01  10/03   SJB  Fix interrupt sharing allocation/free issues.
                       Remove timers as all PCI cards are interrupt-driver
   
   V2.02 09/06    RLH  Fix build problem with 2.4.21-9.EL kernel	
 
   V2.03 02/07    RLH  Enable 16950 mode Fifo's and RTL/TTL interrupt trigger level settings.	       

   V2.04 03/07    RHL  In conjunction with 2.03 allow vmin to set rx fifo.

   V2.05 09/07    RLH  Introduce support for RP550 uPCi 8 & 16 port SMPTE (RS422)card
*********************************************************************************

 * This module exports the following rs232 io functions:
 *
 *	int rs_init(void);
 */

#define	RCKT_DEBUG 0
#define RCKT_DEBUG_MORE 0


/*
 * This activates the fixes by Dr L. Zdunek at EgO Computers Ltd
 * Change EGO_FIX_1 to 0 for to compile for kernels earlier than Linux 2.4
 */


/*V1.02 - makes EGO_FIX_1 selectable via kernel version. 1 = kernel 2.4.0 or greater, 0 = 2.2.x kernels*/
/*#define EGO_FIX_1 (1)*/
#ifndef LINUX_VERSION_CODE
	#include <linux/version.h>
#endif

#ifndef VERSION_CODE
	//#define VERSION_CODE(vers,rel,seq) ( ((vers)<<16)|((rel)<<8)|(seq))
	#define VERSION_CODE(vers,rel,seq) ( ((vers)*65536)+((rel)*256)+(seq))
#endif

#if LINUX_VERSION_CODE<VERSION_CODE(2,4,0)
	#define EGO_FIX_1 (0)
#else
	#define EGO_FIX_1 (1)
#endif
/*V1.02*/


#if EGO_FIX_1
#define XMIT_HEAD(info) ((info)->xmit.head)
#define XMIT_TAIL(info) ((info)->xmit.tail)
#define XMIT_BUF(info)  ((info)->xmit.buf)
#define XMIT_CNT(info) (CIRC_CNT ((info)->xmit.head, (info)->xmit.tail, SERIAL_XMIT_SIZE))
#else
#define XMIT_HEAD(info) ((info)->xmit_head)
#define XMIT_TAIL(info) ((info)->xmit_tail)
#define XMIT_BUF(info) ((info)->xmit_buf)
#define XMIT_CNT(info) (((info)->xmit_head - (info)->xmit_tail) & (SERIAL_XMIT_SIZE - 1))
#endif

#if RCKT_DEBUG
#define tprint if (1) printk
#else
#ifdef DEBUG
#define tprint if (1) printk
#else
#define tprint if (0) printk 
#endif
#endif


/*
 * Serial driver configuration section.  Here are the various options:
 *
 * SERIAL_PARANOIA_CHECK
 * 		Check the magic number for the async_structure where
 * 		ever possible.
 */
#include <linux/config.h>

#undef CONFIG_SERIAL_CONSOLE


#undef SERIAL_PARANOIA_CHECK
#define CONFIG_SERIAL_NOPAUSE_IO
#define SERIAL_DO_RESTART


/* Set of debugging defines */

#if	RCKT_DEBUG
#define SERIAL_DEBUG_INTR
#define SERIAL_DEBUG_IRQ
#define SERIAL_DEBUG_OPEN
#define SERIAL_DEBUG_WRITE
#define DEBUG_PCI
#else
#undef SERIAL_DEBUG_INTR
#undef SERIAL_DEBUG_IRQ
#undef SERIAL_DEBUG_OPEN
#undef SERIAL_DEBUG_WRITE
#undef DEBUG_PCI
#endif
#undef SERIAL_DEBUG_FLOW
#undef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT

#define RS_STROBE_TIME (10*HZ)
#define RS_ISR_PASS_LIMIT 256

#define SERIAL_INLINE

#if defined(MODULE) && defined(SERIAL_DEBUG_MCOUNT)
#define DBG_CNT(s) printk("(%s): [%x] refc=%d, serc=%d, ttyc=%d -> %s\n", \
 kdevname(tty->device), (info->flags), serial_refcount,info->count,tty->count,s)
#else
#define DBG_CNT(s)
#endif


/*
 * End of rckt550 driver configuration section.
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#ifdef SLAB
	#include <linux/slab.h>
#else
	#include <linux/malloc.h>
#endif
#include <linux/init.h>
#include <linux/delay.h>
#ifdef CONFIG_SERIAL_CONSOLE
#include <linux/console.h>
#include <linux/sysrq.h>
#endif

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/serial.h>
#include <asm/atomic.h>
#if	EGO_FIX_1
#include <linux/serialP.h>
#endif

#include <linux/pci.h>

#ifdef SERIAL_INLINE
#define _INLINE_ inline
#endif


MODULE_LICENSE("Proprietary");

#if LINUX_VERSION_CODE > 0x20300
#define PCI_BASE_ADDRESS(dev,num) ((dev)->resource[num].start) & PCI_BASE_ADDRESS_IO_MASK 
#else
#define PCI_BASE_ADDRESS(dev,num) ((dev)->base_address[num]) & 0xfffffffe
#endif

static char *serial_name = "Comtrol Europe Rckt550 driver";
static char *serial_version = "2.05";		

typedef struct
{
	unsigned short int	VendorId;
	unsigned short int	DeviceId;
	unsigned short int	SubDeviceId;
	char                *Description;
	int                  PortBarIndex;
	int                  PortCount;
	int                  Universal;
	int                  SoftCapable;
}
RocketPortDefinition;

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE ~FALSE
#endif


// Original PCI Boards
#define DEVICE_ID_RCKT_550_8RJ11_A	0x8010
#define DEVICE_ID_RCKT_550_8RJ11_B	0x8011
#define DEVICE_ID_RCKT_550_8OCTA_A	0x8012
#define DEVICE_ID_RCKT_550_8OCTA_B	0x8013
#define DEVICE_ID_RCKT_550_4RJ45	   0x8014
#define DEVICE_ID_RCKT_550_4QUAD	   0x8015
#define DEVICE_ID_RCKT_550_16_IF_A	0x8016
#define DEVICE_ID_RCKT_550_16_IF_B	0x8017
#define DEVICE_ID_RCKT_550_8_IF_A	0x8018
#define DEVICE_ID_RCKT_550_8_IF_B	0x8019

// Universal PCI Boards
#define DEVICE_ID_RCKT_550U_4_IF  		0x8020
#define DEVICE_ID_RCKT_550U_8_IF  		0x8021
#define DEVICE_ID_RCKT_550U_16_IF 		0x8022
#define DEVICE_ID_RCKT_550U_8_RJ11		0x8023
#define DEVICE_ID_RCKT_550U_8_OCTA		0x8024
#define DEVICE_ID_RCKT_550U_4_QUAD		0x8025
#define DEVICE_ID_RCKT_550U_4_RJ45		0x8026
#define DEVICE_ID_RCKT_550U_4_MINI  	0x8027
#define DEVICE_ID_RCKT_550U_8_RJ11_232	0x8028
#define DEVICE_ID_RCKT_550U_8_OCTA_232	0x8029
#define DEVICE_ID_RCKT_550U_4_RJ45_232	0x802A
#define DEVICE_ID_RCKT_550U_4_MINI_232  0x802B
#define DEVICE_ID_RCKT_550U_16_SMPTE	0x802C
#define DEVICE_ID_RCKT_550U_8_SMPTE		0x802D

RocketPortDefinition BoardDefinitions[] =
{
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_IF,       0,                          "Universal 4-port I/F.",         3,  4, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_IF,       0,                          "Universal 8-port I/F.",         3,  8, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_16_IF,      0,                          "Universal 16-port I/F.",         3, 16, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_RJ11,     0,                          "Universal 8-port RJ11",         3,  8, TRUE,  TRUE  },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_OCTA,     0,                          "Universal 8-port Octa",         3,  8, TRUE,  TRUE  },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_QUAD,     0,                          "Universal 4-port Quad",         3,  4, TRUE,  TRUE  },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_RJ45,     0,                          "Universal 4-port RJ45",         3,  4, TRUE,  TRUE  },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_MINI,     0,                          "Universal 4-port Mini",         3,  4, TRUE,  TRUE  },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_8RJ11_A,     DEVICE_ID_RCKT_550_8RJ11_B, "Original 8-port RJ11",         0,  8, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_8OCTA_A,     DEVICE_ID_RCKT_550_8OCTA_B, "Original 8-port Octa",         0,  8, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_4RJ45,       0,                          "Original 4-port RJ45",         0,  4, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_4QUAD,       0,                          "Original 4-port Quad",         0,  4, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_16_IF_A,     DEVICE_ID_RCKT_550_16_IF_B, "Original 16-port I/F.",         0, 16, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550_8_IF_A,      DEVICE_ID_RCKT_550_8_IF_B,  "Original 8-port I/F.",         0,  8, FALSE, FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_RJ11_232, 0,                          "Universal 8-port RJ11(RS232)",  3,  8, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_OCTA_232, 0,                          "Universal 8-port Octa(RS232)",  3,  8, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_RJ45_232, 0,                          "Universal 4-port RJ45(RS232)",  3,  4, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_4_MINI_232, 0,                          "Universal 4-port Mini(RS232)",  3,  4, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_16_SMPTE,   0,                          "Universal 16-port SMPTE(RS422)",3, 16, TRUE,  FALSE },
   { PCI_VENDOR_ID_RP, DEVICE_ID_RCKT_550U_8_SMPTE,    0,                          "Universal 8-port SMPTE(RS422)", 3,  8, TRUE,  FALSE },
   { 0,                0,                              0,                          "",                               0,  0, FALSE, FALSE }
};

#define ROCKET550_RS232 0x0c
#define ROCKET550_RS422 0x0f
#define ROCKET550_RS485 0x1a


static DECLARE_TASK_QUEUE(tq_serial);

static struct tty_driver serial_driver, callout_driver;
static struct tty_driver RS422_driver, RS485_driver;
static int serial_refcount;

/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

/*
 * IRQ_timeout		- How long the timeout should be for each IRQ
 * 				should be after the IRQ has been active.
 */

static struct async_struct *IRQ_ports[NR_IRQS];
static struct rs_multiport_struct rs_multiport[NR_IRQS];
static int IRQ_timeout[NR_IRQS];
#ifdef CONFIG_SERIAL_CONSOLE
static struct console sercons;
#  ifdef CONFIG_MAGIC_SYSRQ
   static int sysrq_pressed; /* break, really ... */
#  endif
#endif

/*
 * Here we define the default xmit fifo size used for each type of
 * UART
 */
static struct serial_uart_config uart_config[] = {
	{ "unknown", 1, 0 },
	{ "8250", 1, 0 },	/* 1 */
	{ "16450", 1, 0 },	/* 2 */
	{ "16550", 1, 0 },	/* 3 */
	{ "16550A", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 4 */
	{ "cirrus", 1, 0 },	/* 5 */
	{ "ST16650", 1, UART_CLEAR_FIFO | UART_STARTECH },	/* 6 */
#if EGO_FIX_1
	{ "OX16954", 128, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 7 (Called ST16650V2) */
#else
	{ "OX16954", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 7 (Called ST16650V2) */
#endif
	{ "TI16750", 64, UART_CLEAR_FIFO | UART_USE_FIFO},	/* 8 */
#if EGO_FIX_1
	{ "STARTEC", 1, UART_CLEAR_FIFO | UART_STARTECH },	/* 9 */
	{ "OX16954", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 10 */
	{ "16654", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 11 */
	{ "16850", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 12 */
	{ "RSA", 16, UART_CLEAR_FIFO | UART_USE_FIFO },	/* 13 */
#endif
	{ 0, 0}
};

static struct serial_state rckt550_table[] = {
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,0) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,1) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,2) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,3) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,4) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,5) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,6) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,7) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,8) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,9) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,10) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,11) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,12) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,13) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,14) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(0,15) },

	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,0) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,1) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,2) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,3) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,4) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,5) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,6) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,7) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,8) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,9) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,10) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,11) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,12) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,13) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,14) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(1,15) },
  
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,0) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,1) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,2) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,3) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,4) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,5) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,6) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,7) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,8) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,9) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,10) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,11) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,12) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,13) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,14) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(2,15) },
  
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,0) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,1) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,2) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,3) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,4) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,5) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,6) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,7) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,8) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,9) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,10) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,11) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,12) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,13) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,14) },
	{ 0, BASE_BAUD, 0, 0, STD_COM_FLAGS, C_P(3,15) }
};

#define NR_PORTS	(sizeof(rckt550_table)/sizeof(struct serial_state))

#define MAX_PORTS_PER_BOARD	 16
#define RCKT_MAX_BOARDS			  4
#define OX954_PORTS_PER_CHIP  4
#define UART_SPR	7

static unsigned detect_uart_irq (struct serial_state * state);
static void autoconfig(struct serial_state * info);
static void change_speed(struct async_struct *info, struct termios *old);
static void rckt550_wait_until_sent(struct tty_struct *tty, int timeout);
static int  SetPortMode ( struct serial_state *state, int LineMode );
static unsigned char ReadACR ( struct serial_state *stateC );
static void WriteACR ( struct serial_state *state, unsigned char OrBits, unsigned char AndBits );
//static void WriteTTL ( struct serial_state *state, unsigned char val );
//static void WriteRTL ( struct serial_state *state, unsigned char val );

static int    NumberOfDetectedBoards;
typedef struct
{
   unsigned short BoardBase;
   unsigned short RsModeConfigBase;
	int            NumberOfPortsOnBoard;
}
BoardDataStruct;

static BoardDataStruct BoardData[RCKT_MAX_BOARDS];


static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#define DTR_ENABLE_ACTIVE_LOW			0x10
#define OXSEMI_ACR_3                0x08
#define OXSEMI_ACR_4                0x10

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the copy_from_user blocks while swapping in a page,
 * and some other program tries to do a rckt550 write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the rckt550 ports, since it significantly saves
 * memory if large numbers of rckt550 ports are open.
 */
static unsigned char *tmp_buf;

#if (EGO_FIX_1 && defined(DECLARE_MUTEX))
static DECLARE_MUTEX(tmp_buf_sem);
#else
static struct semaphore tmp_buf_sem = MUTEX;
#endif

static inline int serial_paranoia_check(struct async_struct *info,
					kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char *badmagic =
		"Warning: bad magic number for rckt550 struct (%s) in %s\n";
	static const char *badinfo =
		"Warning: null async_struct for (%s) in %s\n";

	if (!info) {
		printk(badinfo, kdevname(device), routine);
		return 1;
	}
	if (info->magic != SERIAL_MAGIC) {
		printk(badmagic, kdevname(device), routine);
		return 1;
	}
#endif
	return 0;
}

static inline unsigned int serial_in(struct async_struct *info, int offset)
{
	return inb(info->port + offset);
}

static inline unsigned int serial_inp(struct async_struct *info, int offset)
{
#ifdef CONFIG_SERIAL_NOPAUSE_IO
	return inb(info->port + offset);
#else
	return inb_p(info->port + offset);
#endif
}

static inline void serial_out(struct async_struct *info, int offset, int value)
{
#if RCKT_DEBUG_MORE
	tprint ( "w(%x,%ld)=%x %c\n", offset, jiffies, value, (value >= 0x20 && value < 0x7F) ? value : ' ');
#endif

	outb(value, info->port+offset);
}

static inline void serial_outp(struct async_struct *info, int offset,
			       int value)
{
#if RCKT_DEBUG_MORE
	tprint ( "w(p%x,%ld)=%x %c\n", offset, jiffies, value, (value >= 0x20 && value < 0x7F) ? value : ' ');
#endif
#ifdef CONFIG_SERIAL_NOPAUSE_IO
	outb(value, info->port+offset);
#else
    	outb_p(value, info->port+offset);
#endif
}

/*
 * ------------------------------------------------------------
 * rckt550_stop() and rckt550_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void rckt550_stop(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rckt550_stop"))
		return;
	
	save_flags(flags); cli();
	if (info->IER & UART_IER_THRI) {
		info->IER &= ~UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Disable (at B (rckt550_stop) %X) THRI)\n", (int) info->port);
#endif
	}
#if	EGO_FIX_1
	else
	{
		int	tmp;
		if ((tmp = serial_in(info, UART_IER)) & UART_IER_THRI)
		{
			serial_out(info, UART_IER, info->IER);
			printk("rckt550_stop: EGO found TxEN ERROR (IER=%02x)\n", tmp);
		}
	}
#endif
	restore_flags(flags);
}

static void rckt550_start(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rckt550_start"))
		return;
	
	save_flags(flags); cli();
	if (XMIT_CNT(info) && XMIT_BUF(info) && !(info->IER & UART_IER_THRI)) {
		info->IER |= UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Enable (at C(rckt550_start) %X) THRI)\n", (int) info->port);
#endif
	}
	restore_flags(flags);
}

/*
 * ----------------------------------------------------------------------
 *
 * Here starts the interrupt handling routines.  All of the following
 * subroutines are declared as inline and are folded into
 * rs_interrupt().  They were separated out for readability's sake.
 *
 * Note: rs_interrupt() is a "fast" interrupt, which means that it
 * runs with interrupts turned off.  People who may want to modify
 * rs_interrupt() should try to keep the interrupt handler as fast as
 * possible.  After you are done making modifications, it is not a bad
 * idea to do:
 *
 * gcc -S -DKERNEL -Wall -Wstrict-prototypes -O6 -fomit-frame-pointer rckt550.c
 *
 * and look at the resulting assemble code in rckt550.s.
 *
 * 				- Ted Ts'o (tytso@mit.edu), 7-Mar-93
 * -----------------------------------------------------------------------
 */

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static _INLINE_ void rs_sched_event(struct async_struct *info,
				  int event)
{
	info->event |= 1 << event;
	queue_task(&info->tqueue, &tq_serial);
	mark_bh(SERIAL_BH);
}

static _INLINE_ void receive_chars(struct async_struct *info,
				 int *status, struct pt_regs * regs)
{
	struct tty_struct *tty = info->tty;
	unsigned char ch;
	int ignored = 0;
	struct	async_icount *icount;

	icount = &info->state->icount;
	do {
		ch = serial_inp(info, UART_RX);
		if (tty->flip.count >= TTY_FLIPBUF_SIZE)
			break;
		*tty->flip.char_buf_ptr = ch;
		icount->rx++;
		
#ifdef SERIAL_DEBUG_INTR
		printk("DR%02x:%02x...", ch, *status);
#endif
		*tty->flip.flag_buf_ptr = 0;
		if (*status & (UART_LSR_BI | UART_LSR_PE |
			       UART_LSR_FE | UART_LSR_OE)) {
			/*
			 * For statistics only
			 */
			if (*status & UART_LSR_BI) {
#if defined(CONFIG_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
				if (info->line == sercons.index) {
					if (!sysrq_pressed) {
						sysrq_pressed = jiffies;
						goto ignore_char;
					}
					sysrq_pressed = 0;
				}
#endif
				*status &= ~(UART_LSR_FE | UART_LSR_PE);
				icount->brk++;
			} else if (*status & UART_LSR_PE)
				icount->parity++;
			else if (*status & UART_LSR_FE)
				icount->frame++;
			if (*status & UART_LSR_OE)
				icount->overrun++;

			/*
			 * Now check to see if character should be
			 * ignored, and mask off conditions which
			 * should be ignored.
			 */
			if (*status & info->ignore_status_mask) {
				if (++ignored > 100)
					break;
				goto ignore_char;
			}
			*status &= info->read_status_mask;
		
			if (*status & (UART_LSR_BI)) {
#ifdef SERIAL_DEBUG_INTR
				printk("handling break....");
#endif
				*tty->flip.flag_buf_ptr = TTY_BREAK;
				if (info->flags & ASYNC_SAK)
					do_SAK(tty);
			} else if (*status & UART_LSR_PE)
				*tty->flip.flag_buf_ptr = TTY_PARITY;
			else if (*status & UART_LSR_FE)
				*tty->flip.flag_buf_ptr = TTY_FRAME;
			if (*status & UART_LSR_OE) {
				/*
				 * Overrun is special, since it's
				 * reported immediately, and doesn't
				 * affect the current character
				 */
				if (tty->flip.count < TTY_FLIPBUF_SIZE) {
					tty->flip.count++;
					tty->flip.flag_buf_ptr++;
					tty->flip.char_buf_ptr++;
					*tty->flip.flag_buf_ptr = TTY_OVERRUN;
				}
			}
		}
#if defined(CONFIG_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
		if (sysrq_pressed && info->line == sercons.index) {
			if (ch != 0 &&
			    time_before(jiffies, sysrq_pressed + HZ*5)) {
				handle_sysrq(ch, regs, NULL, NULL);
				sysrq_pressed = 0;
				goto ignore_char;
			}
			sysrq_pressed = 0;
		}
#endif
		tty->flip.flag_buf_ptr++;
		tty->flip.char_buf_ptr++;
		tty->flip.count++;
	ignore_char:
		*status = serial_inp(info, UART_LSR);
	} while (*status & UART_LSR_DR);
	tty_flip_buffer_push(tty);
}

static _INLINE_ void transmit_chars(struct async_struct *info, int *intr_done)
{
	int count;
	
	if (info->x_char) {
		serial_outp(info, UART_TX, info->x_char);
		info->state->icount.tx++;
		info->x_char = 0;
		if (intr_done)
			*intr_done = 0;
		return;
	}
	if ((XMIT_CNT(info) <= 0) || info->tty->stopped ||
	    info->tty->hw_stopped) {
		info->IER &= ~UART_IER_THRI;
		serial_out(info, UART_IER, info->IER);
#ifdef SERIAL_DEBUG_INTR_ALL_ENTRIES
	printk("\n(Disable (at A (transmit_chars-stop/empty) %X) THRI)\n", (int) info->port);
		{
			int tmp;
			if ((tmp = serial_in(info, UART_IER)) & UART_IER_THRI)
				printk ("\n*** EGO WARNING: DUFF CHIP %X ***\n", tmp);
		}
#endif
		return;
	}
	
	count = info->xmit_fifo_size;
#ifdef SERIAL_DEBUG_INTR
	printk("Xmit Fifo size=%d...",count);
#endif
	do {
#ifdef SERIAL_DEBUG_INTR
		printk ( "%02x ",  XMIT_BUF(info)[XMIT_TAIL(info)]);
#endif
		serial_out(info, UART_TX, XMIT_BUF(info)[XMIT_TAIL(info)++]);
		XMIT_TAIL(info) = XMIT_TAIL(info) & (SERIAL_XMIT_SIZE-1);
		info->state->icount.tx++;
#if EGO_FIX_1
		/* There is no need to change any implied count */
#else
		--info->xmit_cnt; /* old way reduces the count */
#endif
		if (XMIT_CNT(info) <= 0)
			break;
	} while (--count > 0);
	
	if (XMIT_CNT(info) < WAKEUP_CHARS)
		rs_sched_event(info, RS_EVENT_WRITE_WAKEUP);

#ifdef SERIAL_DEBUG_INTR
	printk("THRE...");
#endif
	if (intr_done)
		*intr_done = 0;

	if (XMIT_CNT(info) <= 0) {
		info->IER &= ~UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Disable (at D (transmit_chars-EMPTY) %X) THRI)\n", (int) info->port);
#endif
		serial_out(info, UART_IER, info->IER);
	}
}

static _INLINE_ void check_modem_status(struct async_struct *info)
{
	int	status;
	struct	async_icount *icount;
	
	status = serial_in(info, UART_MSR);

	if (status & UART_MSR_ANY_DELTA) {
		icount = &info->state->icount;
		/* update input line counters */
		if (status & UART_MSR_TERI)
			icount->rng++;
		if (status & UART_MSR_DDSR)
			icount->dsr++;
		if (status & UART_MSR_DDCD) {
			icount->dcd++;
#ifdef CONFIG_HARD_PPS
			if ((info->flags & ASYNC_HARDPPS_CD) &&
			    (status & UART_MSR_DCD))
				hardpps();
#endif
		}
		if (status & UART_MSR_DCTS)
			icount->cts++;
		wake_up_interruptible(&info->delta_msr_wait);
	}

	if ((info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD)) {
#if (defined(SERIAL_DEBUG_OPEN) || defined(SERIAL_DEBUG_INTR))
		printk("ttys%d CD now %s...", info->line,
		       (status & UART_MSR_DCD) ? "on" : "off");
#endif		
		if (status & UART_MSR_DCD)
			wake_up_interruptible(&info->open_wait);
		else if (!((info->flags & ASYNC_CALLOUT_ACTIVE) &&
			   (info->flags & ASYNC_CALLOUT_NOHUP))) {
#ifdef SERIAL_DEBUG_OPEN
			printk("doing rckt550 hangup...");
#endif
			if (info->tty)
				tty_hangup(info->tty);
		}
	}
	if (info->flags & ASYNC_CTS_FLOW) {
		if (info->tty->hw_stopped) {
			if (status & UART_MSR_CTS) {
#if (defined(SERIAL_DEBUG_INTR) || defined(SERIAL_DEBUG_FLOW))
				printk("CTS tx start...");
#endif
				info->tty->hw_stopped = 0;
				info->IER |= UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Enable (at E(check_modem_status) %X) THRI)\n", (int) info->port);
#endif
				serial_out(info, UART_IER, info->IER);
				rs_sched_event(info, RS_EVENT_WRITE_WAKEUP);
				return;
			}
		} else {
			if (!(status & UART_MSR_CTS)) {
#if (defined(SERIAL_DEBUG_INTR) || defined(SERIAL_DEBUG_FLOW))
				printk("CTS tx stop...");
#endif
				info->tty->hw_stopped = 1;
				info->IER &= ~UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Disable (at F (check_modem_status) %X) THRI)\n", (int) info->port);
#endif
				serial_out(info, UART_IER, info->IER);
			}
		}
	}
}

/*
 * This is the rckt550 driver's generic interrupt routine
 */
static void rs_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	int status;
	struct async_struct * info;
	int pass_counter = 0;
	struct async_struct *end_mark = 0;
	int first_multi = 0;
	struct rs_multiport_struct *multi;

#ifdef SERIAL_DEBUG_INTR
	printk("rs_interrupt(%d)...", irq);
#endif
	
	info = IRQ_ports[irq];
	if (!info)
		return;
	
	multi = &rs_multiport[irq];
	if (multi->port_monitor)
		first_multi = inb(multi->port_monitor);

	do {
		if (!info->tty ||
		    (serial_in(info, UART_IIR) & UART_IIR_NO_INT)) {
			if (!end_mark)
				end_mark = info;
			goto next;
		}
		end_mark = 0;

		info->last_active = jiffies;

		status = serial_inp(info, UART_LSR);
#ifdef SERIAL_DEBUG_INTR
		printk("status = %x...", status);
#endif
		if (status & UART_LSR_DR)
			receive_chars(info, &status, regs);
		check_modem_status(info);
		if (status & UART_LSR_THRE)
			transmit_chars(info, 0);

	next:
		info = info->next_port;
		if (!info) {
			info = IRQ_ports[irq];
			if (pass_counter++ > RS_ISR_PASS_LIMIT) {
#if 0
				printk("rs loop break\n");
#endif
				break; 	/* Prevent infinite loops */
			}
			continue;
		}
	} while (end_mark != info);
	if (multi->port_monitor)
		printk("rs port monitor (normal) irq %d: 0x%x, 0x%x\n",
		       info->state->irq, first_multi,
		       inb(multi->port_monitor));
#ifdef SERIAL_DEBUG_INTR
	printk("end.\n");
#endif
}


/*
 * This is the rckt550 driver's interrupt routine for a single port
 */
static void rs_interrupt_single(int irq, void *dev_id, struct pt_regs * regs)
{
	int status;
	int pass_counter = 0;
	struct async_struct * info;
	int first_multi = 0;
	struct rs_multiport_struct *multi;
	
#ifdef SERIAL_DEBUG_INTR_ALL_ENTRIES
	printk("rs_interrupt_single(%d) at %ld...", irq, jiffies);
#endif
	
	info = IRQ_ports[irq];
	if (!info || !info->tty)
		return;

	multi = &rs_multiport[irq];
	if (multi->port_monitor)
		first_multi = inb(multi->port_monitor);

	do {
		status = serial_inp(info, UART_LSR);
#ifdef SERIAL_DEBUG_INTR_ALL_ENTRIES
		{
		int status1;
		status1 = serial_inp(info, UART_IER);
		printk("status(%X)= %X,%X...", (int) info->port, status, status1);
		}
#endif
		if (status & UART_LSR_DR)
			receive_chars(info, &status, regs);
		check_modem_status(info);
		if (status & UART_LSR_THRE)
			transmit_chars(info, 0);
		if (pass_counter++ > RS_ISR_PASS_LIMIT) {
#ifdef SERIAL_DEBUG_INTR
			printk("rs_single loop break.\n");
#endif
			break;
		}
#if RCKT_DEBUG_MORE
			printk("Done a bunch of RX/TX at %ld\n", jiffies);
#endif
	} while (!(serial_in(info, UART_IIR) & UART_IIR_NO_INT));
	info->last_active = jiffies;
	if (multi->port_monitor)
		printk("rs port monitor (single) irq %d: 0x%x, 0x%x\n",
		       info->state->irq, first_multi,
		       inb(multi->port_monitor));
#ifdef SERIAL_DEBUG_INTR_ALL_ENTRIES
	printk("end at %ld.\n", jiffies);
#endif
}

/*
 * This is the rckt550 driver's for multiport boards
 */
static void rs_interrupt_multi(int irq, void *dev_id, struct pt_regs * regs)
{
	int status;
	struct async_struct * info;
	int pass_counter = 0;
	int first_multi= 0;
	struct rs_multiport_struct *multi;

#ifdef SERIAL_DEBUG_INTR
	printk("rs_interrupt_multi(%d)...", irq);
#endif
	
	info = IRQ_ports[irq];
	if (!info)
		return;
	multi = &rs_multiport[irq];
	if (!multi->port1) {
		/* Should never happen */
		printk("rs_interrupt_multi: NULL port1!\n");
		return;
	}
	if (multi->port_monitor)
		first_multi = inb(multi->port_monitor);
	
	while (1) {
		if (!info->tty ||
		    (serial_in(info, UART_IIR) & UART_IIR_NO_INT))
			goto next;

		info->last_active = jiffies;

		status = serial_inp(info, UART_LSR);
#ifdef SERIAL_DEBUG_INTR
		printk("status = %x...", status);
#endif
		if (status & UART_LSR_DR)
			receive_chars(info, &status, regs);
		check_modem_status(info);
		if (status & UART_LSR_THRE)
			transmit_chars(info, 0);

	next:
		info = info->next_port;
		if (info)
			continue;

		info = IRQ_ports[irq];
		if (pass_counter++ > RS_ISR_PASS_LIMIT) {
#if 1
			printk("rs_multi loop break\n");
#endif
			break; 	/* Prevent infinite loops */
		}
		if (multi->port_monitor)
			printk("rs port monitor irq %d: 0x%x, 0x%x\n",
			       info->state->irq, first_multi,
			       inb(multi->port_monitor));
		if ((inb(multi->port1) & multi->mask1) != multi->match1)
			continue;
		if (!multi->port2)
			break;
		if ((inb(multi->port2) & multi->mask2) != multi->match2)
			continue;
		if (!multi->port3)
			break;
		if ((inb(multi->port3) & multi->mask3) != multi->match3)
			continue;
		if (!multi->port4)
			break;
		if ((inb(multi->port4) & multi->mask4) != multi->match4)
			continue;
		break;
	}
#ifdef SERIAL_DEBUG_INTR
	printk("end.\n");
#endif
}

/*
 * -------------------------------------------------------------------
 * Here ends the rckt550 interrupt routines.
 * -------------------------------------------------------------------
 */

/*
 * This routine is used to handle the "bottom half" processing for the
 * rckt550 driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * rs_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using rs_sched_event(), and they get done here.
 */
static void do_serial_bh(void)
{
	run_task_queue(&tq_serial);
}

static void do_softint(void *private_)
{
	struct async_struct	*info = (struct async_struct *) private_;
	struct tty_struct	*tty;
	
	tty = info->tty;
	if (!tty)
		return;

	if (test_and_clear_bit(RS_EVENT_WRITE_WAKEUP, &info->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
#ifdef LINUX_7
		wake_up_interruptible(&tty->poll_wait);
#endif
	}
}


/*
 * ---------------------------------------------------------------
 * Low level utility subroutines for the rckt550 driver:  routines to
 * figure out the appropriate timeout for an interrupt chain, routines
 * to initialize and startup a rckt550 port, and routines to shutdown a
 * rckt550 port.  Useful stuff like that.
 * ---------------------------------------------------------------
 */

/*
 * This routine figures out the correct timeout for a particular IRQ.
 * It uses the smallest timeout of all of the rckt550 ports in a
 * particular interrupt chain.  Now only used for IRQ 0....
 */
static void figure_IRQ_timeout(int irq)
{
	struct	async_struct	*info;
	int	timeout = 60*HZ;	/* 60 seconds === a long time :-) */

	info = IRQ_ports[irq];
	if (!info) {
		IRQ_timeout[irq] = 60*HZ;
		return;
	}
	while (info) {
		if (info->timeout < timeout)
			timeout = info->timeout;
		info = info->next_port;
	}
	if (!irq)
		timeout = timeout / 2;
	IRQ_timeout[irq] = timeout ? timeout : 1;
}

/*
 * This routine is called to set the UART FIFO registers to match
 * VMIN for the OX16954 chips - REL2.04.
 */
static void set_fifo(struct async_struct *info)
{
	int vmin = info->tty->termios->c_cc[VMIN];

	serial_outp(info, UART_LCR, 0xBF);
	if (serial_in(info, UART_EFR) == 0)
	{
		/*enable 16650 mode REL2.04 modification for DCS*/
		/*switch on enhanced mode*/
		unsigned char efrval = serial_in(info, UART_EFR) | UART_EFR_ECB;
		serial_outp(info,UART_EFR,efrval);
	}
	serial_outp( info, UART_SPR, 0 );   /*Select ACR register as destination for next ICR write*/
	serial_outp( info, UART_ICR, 0x20); /*enable enhanced mode fifo trigger levels*/
	serial_outp( info, UART_SPR, 4 );   /*Select TX Fifo trigger level register*/
	serial_outp( info, UART_ICR, 1 );   /*Initialise to 1*/
	serial_outp( info, UART_SPR, 5 );   /*Select RX Fifo trigger level register*/
	if (vmin > 1 && vmin <= 64) {
		serial_outp( info, UART_ICR, vmin );	/*Initialise to vmin*/
	} else {
		serial_outp( info, UART_ICR, 1 );	/*Initialise to 1*/
	}
	serial_outp(info, UART_LCR, 0);
}

static int startup(struct async_struct * info)
{
	unsigned long flags;
	int	retval=0;
	void (*handler)(int, void *, struct pt_regs *);
	struct serial_state *state= info->state;
	unsigned long page;
	unsigned short ICP;

	page = get_free_page(GFP_KERNEL);
	if (!page)
		return -ENOMEM;

	save_flags(flags); cli();

	if (info->flags & ASYNC_INITIALIZED)
	{
		free_page(page);
		goto errout;
	}

	if (!state->port || !state->type)
	{
		if (info->tty)
			set_bit(TTY_IO_ERROR, &info->tty->flags);
		free_page(page);
		goto errout;
	}
	if (XMIT_BUF(info))
		free_page(page);
	else
		XMIT_BUF(info) = (unsigned char *) page;

#ifdef SERIAL_DEBUG_OPEN
	printk("startup: starting up ttys%d (irq %d) (type %d)...\n", info->line, state->irq, info->state->type);
#endif

	if (info->state->type == PORT_16650V2) {
		set_fifo(info);
	}

	if (uart_config[info->state->type].flags & UART_STARTECH)
	{
		/* Wake up UART */
		serial_outp(info, UART_LCR, 0xBF);
		serial_outp(info, UART_EFR, UART_EFR_ECB);
		serial_outp(info, UART_IER, 0);
		serial_outp(info, UART_EFR, 0);
		serial_outp(info, UART_LCR, 0);
	}

	if (info->state->type == PORT_16750)
	{
		/* Wake up UART */
		serial_outp(info, UART_IER, 0);
	}


	/*
	 * Clear the FIFO buffers and disable them
	 * (they will be reenabled in change_speed())
	 */
	if (uart_config[state->type].flags & UART_CLEAR_FIFO)
		serial_outp(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
					     UART_FCR_CLEAR_XMIT));

	/*
	 * At this point there's no way the LSR could still be 0xFF;
	 * if it is, then bail out, because there's likely no UART
	 * here.
	 */
	if (serial_inp(info, UART_LSR) == 0xff)
	{
		if (capable(CAP_SYS_ADMIN))
		{
			if (info->tty)
				set_bit(TTY_IO_ERROR, &info->tty->flags);
		} else
			retval = -ENODEV;
		goto errout;
	}
	
	/*
	 * Allocate the IRQ if necessary
	 */
	if (state->irq && (!IRQ_ports[state->irq] || !IRQ_ports[state->irq]->next_port)) 
	{
		if (IRQ_ports[state->irq]) 
		{
#ifdef SERIAL_DEBUG_IRQ
			printk ( "in startup: Freeing IRQ%d\n", state->irq );
#endif
			
			free_irq(state->irq, serial_name);
			
			if (rs_multiport[state->irq].port1)
				handler = rs_interrupt_multi;
			else
				handler = rs_interrupt;
		}
	        else
			handler = rs_interrupt_single;

		/*
		 *  All our interrupts are shared (because all cards are PCI), so we pass in 
		 *  the ASYNC_STRUC address as a cookie that helps the kernel manage freeing
		 *  at some later time.
		 */
#ifdef SERIAL_DEBUG_IRQ
		printk ( "in startup : Requesting IRQ%d\n", state->irq );
#endif
		retval = request_irq(state->irq, handler, SA_SHIRQ,
				     "rckt550", serial_name);
		if (retval)
		{
			printk ("EGO: COULD NOT ADD INTERRUPT HANDLER: (ttys%d) irq=%d,retval=%d,(%s)\n",
								 info->line, state->irq, retval, "SHARED" );
			if (capable(CAP_SYS_ADMIN)) 
			{
				if (info->tty)
					set_bit(TTY_IO_ERROR, &info->tty->flags);
				retval = 0;
			}
			goto errout;
		}
	}

	/*
	 * Insert rckt550 port into IRQ chain.
	 */
	info->prev_port = 0;
	info->next_port = IRQ_ports[state->irq];
	if (info->next_port)
		info->next_port->prev_port = info;
	IRQ_ports[state->irq] = info;
	figure_IRQ_timeout(state->irq);

	/*
	 * Clear the interrupt registers.
	 */
     /* (void) serial_inp(info, UART_LSR); */   /* (see above) */
	(void) serial_inp(info, UART_RX);
	(void) serial_inp(info, UART_IIR);
	(void) serial_inp(info, UART_MSR);

	/*
	 * Now, initialize the UART
	 */
	serial_outp(info, UART_LCR, UART_LCR_WLEN8);	/* reset DLAB */

	info->MCR = 0;
	if (info->tty->termios->c_cflag & CBAUD)
		info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	if (info->flags & ASYNC_FOURPORT)
	{
		if (state->irq == 0)
			info->MCR |= UART_MCR_OUT1;
	}
	else
	{
		if (state->irq != 0)
			info->MCR |= UART_MCR_OUT2;
	}
	serial_outp(info, UART_MCR, info->MCR);
	
	/*
	 * Finally, enable interrupts
	 */
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;
	serial_outp(info, UART_IER, info->IER);	/* enable interrupts */
	
	if (info->flags & ASYNC_FOURPORT)
	{
		/* Enable interrupts on the AST Fourport board */
		ICP = (info->port & 0xFE0) | 0x01F;
		outb_p(0x80, ICP);
		(void) inb_p(ICP);
	}

	/*
	 * And clear the interrupt registers again for luck.
	 */
	(void)serial_inp(info, UART_LSR);
	(void)serial_inp(info, UART_RX);
	(void)serial_inp(info, UART_IIR);
	(void)serial_inp(info, UART_MSR);

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	XMIT_HEAD(info) = XMIT_TAIL(info) = 0;
#if EGO_FIX_1
	/* The implied count is automatically 0 */
#else
	info->xmit_cnt = 0;
#endif


	/*
	 * Set up the tty->alt_speed kludge
	 */
	if (info->tty)
	{
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			info->tty->alt_speed = 57600;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			info->tty->alt_speed = 115200;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			info->tty->alt_speed = 230400;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			info->tty->alt_speed = 460800;
	}
	
	/*
	 * and set the speed of the rckt550 port
	 */
	change_speed(info, 0);

	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
#if RCKT_DEBUG
	printk("EGO: startup done OK (A)\n");
#endif
	return 0;
	
errout:
	restore_flags(flags);
	if (retval)
		printk("EGO: startup FAILED (A)\n");
	return retval;
}

/*
 * This routine will shutdown a rckt550 port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct async_struct * info)
{
	unsigned long	flags;
	struct serial_state *state;
	int		retval;

	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	state = info->state;

#ifdef SERIAL_DEBUG_OPEN
	printk("Shutting down rckt550 port %d (irq %d)....", info->line,
	       state->irq);
#endif
	
	save_flags(flags); cli(); /* Disable interrupts */

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);
	
	/*
	 * First unlink the rckt550 port from the IRQ chain...
	 */
	if (info->next_port)
		info->next_port->prev_port = info->prev_port;
	if (info->prev_port)
		info->prev_port->next_port = info->next_port;
	else
		IRQ_ports[state->irq] = info->next_port;
	figure_IRQ_timeout(state->irq);
	
	/*
	 * Free the IRQ, if necessary
	 */
	if (state->irq && (!IRQ_ports[state->irq] ||
			  !IRQ_ports[state->irq]->next_port)) {
#ifdef SERIAL_DEBUG_IRQ
		printk ( "in shutdown(1) Have IRQ%d\n", state->irq );
#endif
		if (IRQ_ports[state->irq])
		 {
#ifdef SERIAL_DEBUG_IRQ
			printk ( "in shutdown(1) Freeing IRQ%d\n", state->irq );
#endif
			free_irq(state->irq, serial_name);
#ifdef SERIAL_DEBUG_IRQ
			printk ( "in shutdown(1) Requesting IRQ%d\n", state->irq );
#endif
			retval = request_irq(state->irq, rs_interrupt_single,
					     SA_SHIRQ, "rckt550", serial_name);
			
			if (retval)
				printk("rckt550 shutdown: request_irq: error %d"
				       "  Couldn't reacquire IRQ.\n", retval);
		}
		else
		{
#ifdef SERIAL_DEBUG_IRQ
			printk ( "in shutdown(2) Freeing IRQ%d\n", state->irq );
#endif
			free_irq(state->irq, serial_name);
		}
	}

	if (XMIT_BUF(info)) {
		free_page((unsigned long) XMIT_BUF(info));
		XMIT_BUF(info) = 0;
	}

	info->IER = 0;
	serial_outp(info, UART_IER, 0x00);	/* disable all intrs */
	if (info->flags & ASYNC_FOURPORT) {
		/* reset interrupts on the AST Fourport board */
		(void) inb((info->port & 0xFE0) | 0x01F);
		info->MCR |= UART_MCR_OUT1;
	} else
		info->MCR &= ~UART_MCR_OUT2;
	
	/* disable break condition */
	serial_out(info, UART_LCR, serial_inp(info, UART_LCR) & ~UART_LCR_SBC);
	
	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);
	serial_outp(info, UART_MCR, info->MCR);

	/* disable FIFO's */	
	serial_outp(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
				     UART_FCR_CLEAR_XMIT));
	(void)serial_in(info, UART_RX);    /* read data port to reset things */
	
	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);

	if (uart_config[info->state->type].flags & UART_STARTECH) {
		/* Arrange to enter sleep mode */
		serial_outp(info, UART_LCR, 0xBF);
		serial_outp(info, UART_EFR, UART_EFR_ECB);
		serial_outp(info, UART_IER, UART_IERX_SLEEP);
		serial_outp(info, UART_LCR, 0);
	}
	if (info->state->type == PORT_16750) {
		/* Arrange to enter sleep mode */
		serial_outp(info, UART_IER, UART_IERX_SLEEP);
	}
	SetPortMode(state,0);								/* TRI-State the line drivers on soft cards */
	info->flags &= ~ASYNC_INITIALIZED;

	restore_flags(flags);
}

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a rckt550 port.
 */
static void change_speed(struct async_struct *info,
			 struct termios *old_termios)
{
	unsigned short port;
	int	quot = 0, baud_base, baud;
	unsigned cflag, cval, fcr = 0;
	int	bits;
	unsigned long	flags;

	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;
	if (!(port = info->port))
		return;

	/* byte size and parity */
	switch (cflag & CSIZE) {
	      case CS5: cval = 0x00; bits = 7; break;
	      case CS6: cval = 0x01; bits = 8; break;
	      case CS7: cval = 0x02; bits = 9; break;
	      case CS8: cval = 0x03; bits = 10; break;
	      /* Never happens, but GCC is too dumb to figure it out */
	      default:  cval = 0x00; bits = 7; break;
	      }
	if (cflag & CSTOPB) {
		cval |= 0x04;
		bits++;
	}
	if (cflag & PARENB) {
		cval |= UART_LCR_PARITY;
		bits++;
	}
	if (!(cflag & PARODD))
		cval |= UART_LCR_EPAR;
#ifdef CMSPAR
	if (cflag & CMSPAR)
		cval |= UART_LCR_SPAR;
#endif

	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(info->tty);
	if (!baud)
		baud = 9600;	/* B0 transition handled in rckt550_set_termios */
	baud_base = info->state->baud_base;

   tprint ( "Baud Base = %d\n", baud_base );

	if (baud == 38400 &&
	    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
		quot = info->state->custom_divisor;
	else {
		if (baud == 134)
			/* Special case since 134 is really 134.5 */
			quot = (2*baud_base / 269);
		else if (baud)
			quot = baud_base / baud;
	}

	tprint ( "Speed:cflag=%x, baud=%d, quot=%x\n", cflag, baud, quot );

	/* If the quotient is zero refuse the change */
	if (!quot && old_termios) {
		info->tty->termios->c_cflag &= ~CBAUD;
		info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
		baud = tty_get_baud_rate(info->tty);
		if (!baud)
			baud = 9600;
		if (baud == 38400 &&
		    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
			quot = info->state->custom_divisor;
		else {
			if (baud == 134)
				/* Special case since 134 is really 134.5 */
				quot = (2*baud_base / 269);
			else if (baud)
				quot = baud_base / baud;
		}
	}
	/* As a last resort, if the quotient is zero, default to 9600 bps */
	if (!quot)
		quot = baud_base / 9600;
	info->quot = quot;
	info->timeout = ((info->xmit_fifo_size*HZ*bits*quot) / baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	/* Set up FIFO's */
	/*
		Rel 2.03 DCS mod. The uart is configured into enhanced 950 mode by at init time
		Fifo trigger levels are set in the RTL and TTL registers which reside in the enhanced mode
		ICR register set. In enhanced mode the FCR register fifo trigger level bits are ignored.
		So UART_FCR_TRIGGER_1 and UART_FCR_TRIGGER_8 are ignored.
	*/
	
	if (uart_config[info->state->type].flags & UART_USE_FIFO) {
		if ((info->state->baud_base / quot) < 2400)
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_1;
		else
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_8;
	}
	if (info->state->type == PORT_16750)
		fcr |= UART_FCR7_64BYTE;
	
	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;
	if (info->flags & ASYNC_HARDPPS_CD)
		info->IER |= UART_IER_MSI;
	if (cflag & CRTSCTS) {
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
	} else
		info->flags &= ~ASYNC_CTS_FLOW;
	if (cflag & CLOCAL)
		info->flags &= ~ASYNC_CHECK_CD;
	else {
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}
	serial_out(info, UART_IER, info->IER);

	/*
	 * Set up parity check flag
	 */
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))

	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if (I_INPCK(info->tty))
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
		info->read_status_mask |= UART_LSR_BI;
	
#if defined(CONFIG_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
	if (info->line == sercons.index)
		info->read_status_mask |= UART_LSR_BI;
#endif
	/*
	 * Characters to ignore
	 */
	info->ignore_status_mask = 0;
	if (I_IGNPAR(info->tty))
		info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	if (I_IGNBRK(info->tty)) {
		info->ignore_status_mask |= UART_LSR_BI;
		/*
		 * If we're ignore parity and break indicators, ignore
		 * overruns too.  (For real raw support).
		 */
		if (I_IGNPAR(info->tty))
			info->ignore_status_mask |= UART_LSR_OE;
	}
	/*
	 * !!! ignore all characters if CREAD is not set
	 */
	if ((cflag & CREAD) == 0)
		info->ignore_status_mask |= UART_LSR_DR;
	save_flags(flags); cli();
	if (uart_config[info->state->type].flags & UART_STARTECH) {
		serial_outp(info, UART_LCR, 0xBF);
		serial_outp(info, UART_EFR,
			    (cflag & CRTSCTS) ? UART_EFR_CTS : 0);
	}
	if (info->state->type == PORT_16650V2) {
		set_fifo(info);
	}
	serial_outp(info, UART_LCR, cval | UART_LCR_DLAB);	/* set DLAB */
	serial_outp(info, UART_DLL, quot & 0xff);	/* LS of divisor */
	serial_outp(info, UART_DLM, quot >> 8);		/* MS of divisor */
	if (info->state->type == PORT_16750)
		serial_outp(info, UART_FCR, fcr); 	/* set fcr */
	serial_outp(info, UART_LCR, cval);		/* reset DLAB */
	if (info->state->type != PORT_16750) {
		if (fcr & UART_FCR_ENABLE_FIFO) {
			/* emulated UARTs (Lucent Venus 167x) need two steps */
			serial_outp(info, UART_FCR, UART_FCR_ENABLE_FIFO);
		}
		serial_outp(info, UART_FCR, fcr); 	/* set fcr */
	}

	restore_flags(flags);
}

static void rckt550_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rckt550_put_char"))
		return;

	if (!tty || !XMIT_BUF(info))
		return;

	save_flags(flags); cli();
	if (XMIT_CNT(info) >= SERIAL_XMIT_SIZE - 1) {
		restore_flags(flags);
		return;
	}

	XMIT_BUF(info)[XMIT_HEAD(info)++] = ch;
	XMIT_HEAD(info) &= SERIAL_XMIT_SIZE-1;
#if EGO_FIX_1
	/* The implied count has increased */
#else
	info->xmit_cnt++;
#endif
	restore_flags(flags);
}

static void rckt550_flush_chars(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
				
	if (serial_paranoia_check(info, tty->device, "rckt550_flush_chars"))
		return;

	if (XMIT_CNT(info) <= 0 || tty->stopped || tty->hw_stopped ||
	    !XMIT_BUF(info))
		return;

	save_flags(flags); cli();
	info->IER |= UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Enable (at H (flush_chars) %X) THRI)\n", (int) info->port);
#endif
	serial_out(info, UART_IER, info->IER);
	restore_flags(flags);
}

static int rckt550_write(struct tty_struct * tty, int from_user,
		    const unsigned char *buf, int count)
{
	int	c, ret = 0;
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
				
#ifdef SERIAL_DEBUG_WRITE
	printk ( "rckt550_write time=%ld(%d chars): ", jiffies, count );
	for ( c=0 ; c<count ; c++ )
	{
	   printk ( "%02x ", buf[c] );
	}
	printk ( "\n" );
#endif
	
	if (serial_paranoia_check(info, tty->device, "rckt550_write"))
		return 0;

	if (!tty || !XMIT_BUF(info) || !tmp_buf)
		return 0;

	save_flags(flags);
	if (from_user) {
#ifdef SERIAL_DEBUG_WRITE
	   printk ( "from_user\n" );
#endif
		down(&tmp_buf_sem);
		while (1) {
			c = MIN(count,
				MIN(SERIAL_XMIT_SIZE - XMIT_CNT(info) - 1,
				    SERIAL_XMIT_SIZE - XMIT_HEAD(info)));
			if (c <= 0)
				break;

			c -= copy_from_user(tmp_buf, buf, c);
			if (!c) {
				if (!ret)
					ret = -EFAULT;
				break;
			}
			cli();
			c = MIN(c, MIN(SERIAL_XMIT_SIZE - XMIT_CNT(info) - 1,
				       SERIAL_XMIT_SIZE - XMIT_HEAD(info)));
			memcpy(XMIT_BUF(info) + XMIT_HEAD(info), tmp_buf, c);
			XMIT_HEAD(info) = ((XMIT_HEAD(info) + c) &
					   (SERIAL_XMIT_SIZE-1));
#if EGO_FIX_1
			/* The implied count has changed */
#else
			info->xmit_cnt += c;
#endif
			restore_flags(flags);
			buf += c;
			count -= c;
			ret += c;
		}
		up(&tmp_buf_sem);
	} else {
#ifdef SERIAL_DEBUG_WRITE
	   printk ( "from_system\n" );
#endif
		while (1) {
			cli();		
			c = MIN(count,
				MIN(SERIAL_XMIT_SIZE - XMIT_CNT(info) - 1,
				    SERIAL_XMIT_SIZE - XMIT_HEAD(info)));
			if (c <= 0) {
				restore_flags(flags);
				break;
			}
#ifdef SERIAL_DEBUG_WRITE
	      printk ( "copying %d bytes to %lx\n", c, (long)(XMIT_BUF(info) + XMIT_HEAD(info)) );
#endif
			memcpy(XMIT_BUF(info) + XMIT_HEAD(info), buf, c);
			XMIT_HEAD(info) = ((XMIT_HEAD(info) + c) &
					   (SERIAL_XMIT_SIZE-1));
#if EGO_FIX_1
			/* The implied count has changed */
#else
			info->xmit_cnt += c;
#endif
			restore_flags(flags);
			buf += c;
			count -= c;
			ret += c;
		}
	}
	if (XMIT_CNT(info) && !tty->stopped && !tty->hw_stopped &&
	    !(info->IER & UART_IER_THRI)) {
		info->IER |= UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Enable (at J (rckt550_write) %X) THRI)\n", (int) info->port);
#endif
		serial_out(info, UART_IER, info->IER);
	}
	return ret;
}

static int rckt550_write_room(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	int	ret;
				
	if (serial_paranoia_check(info, tty->device, "rckt550_write_room"))
		return 0;
	ret = SERIAL_XMIT_SIZE - XMIT_CNT(info) - 1;
	if (ret < 0)
		ret = 0;
	return ret;
}

static int rckt550_chars_in_buffer(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "rckt550_chars_in_buffer"))
		return 0;
	return XMIT_CNT(info);
}

static void rckt550_flush_buffer(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "rckt550_flush_buffer"))
		return;
	save_flags(flags); cli();
	XMIT_HEAD(info) = XMIT_TAIL(info) = 0;
#if EGO_FIX_1
	/* The implied count has changed */
#else
	info->xmit_cnt = 0;
#endif
	restore_flags(flags);
	wake_up_interruptible(&tty->write_wait);
#ifdef LINUX_7
	wake_up_interruptible(&tty->poll_wait);
#endif
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*
 * This function is used to send a high-priority XON/XOFF character to
 * the device
 */
static void rckt550_send_xchar(struct tty_struct *tty, char ch)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;

	if (serial_paranoia_check(info, tty->device, "rs_send_char"))
		return;

	info->x_char = ch;
	if (ch) {
		/* Make sure transmit interrupts are on */
		info->IER |= UART_IER_THRI;
#ifdef SERIAL_DEBUG_INTR
	printk("\n(Enable (at K (send_xchar) %X) THRI)\n", (int) info->port);
#endif
		serial_out(info, UART_IER, info->IER);
	}
}

/*
 * ------------------------------------------------------------
 * rckt550_throttle()
 *
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */
static void rckt550_throttle(struct tty_struct * tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("throttle %s: %d....\n", tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rckt550_throttle"))
		return;
	
	if (I_IXOFF(tty))
		rckt550_send_xchar(tty, STOP_CHAR(tty));

	if (tty->termios->c_cflag & CRTSCTS)
		info->MCR &= ~UART_MCR_RTS;

	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
}

static void rckt550_unthrottle(struct tty_struct * tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("unthrottle %s: %d....\n", tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rckt550_unthrottle"))
		return;
	
	if (I_IXOFF(tty)) {
		if (info->x_char)
			info->x_char = 0;
		else
			rckt550_send_xchar(tty, START_CHAR(tty));
	}
	if (tty->termios->c_cflag & CRTSCTS)
		info->MCR |= UART_MCR_RTS;
	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
}

/*
 * ------------------------------------------------------------
 * rckt550_ioctl() and friends
 * ------------------------------------------------------------
 */

static int get_serial_info(struct async_struct * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	struct serial_state *state = info->state;

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = state->type;
	tmp.line = state->line;
	tmp.port = state->port;
	tmp.irq = state->irq;
	tmp.flags = state->flags;
	tmp.xmit_fifo_size = state->xmit_fifo_size;
	tmp.baud_base = state->baud_base;
	tmp.close_delay = state->close_delay;
	tmp.closing_wait = state->closing_wait;
	tmp.custom_divisor = state->custom_divisor;
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_serial_info(struct async_struct * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
 	struct serial_state old_state, *state;
	unsigned int		i,change_irq,change_port;
	int 			retval = 0;

	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
		return -EFAULT;
	state = info->state;
	old_state = *state;

#if RCKT_DEBUG_MORE
	printk (KERN_INFO "ENTER set_serial_info with struct at %X has type=%d\n", (int)info, state ? state->type : -1);
#endif
	change_irq = new_serial.irq != state->irq;
	change_port = (new_serial.port != state->port);

	if (!capable(CAP_SYS_ADMIN)) {
		if (change_irq || change_port ||
		    (new_serial.baud_base != state->baud_base) ||
		    (new_serial.type != state->type) ||
		    (new_serial.close_delay != state->close_delay) ||
		    (new_serial.xmit_fifo_size != state->xmit_fifo_size) ||
		    ((new_serial.flags & ~ASYNC_USR_MASK) !=
		     (state->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		state->flags = ((state->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		state->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	new_serial.irq = irq_cannonicalize(new_serial.irq);

	if ((new_serial.irq >= NR_IRQS) || (new_serial.port > 0xffff) ||
	    (new_serial.baud_base < 9600)|| (new_serial.type < PORT_UNKNOWN) ||
	    (new_serial.type > PORT_MAX) || (new_serial.type == PORT_CIRRUS) ||
	    (new_serial.type == PORT_STARTECH)) {
		return -EINVAL;
	}

	if ((new_serial.type != state->type) ||
	    (new_serial.xmit_fifo_size <= 0))
		new_serial.xmit_fifo_size =
			uart_config[new_serial.type].dfl_xmit_fifo_size;

	/* Make sure address is not already in use */
	if (new_serial.type) {
		for (i = 0 ; i < NR_PORTS; i++)
			if ((state != &rckt550_table[i]) &&
			    (rckt550_table[i].port == new_serial.port) &&
			    rckt550_table[i].type)
				return -EADDRINUSE;
	}

	if ((change_port || change_irq) && (state->count > 1))
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	state->baud_base = new_serial.baud_base;
	state->flags = ((state->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->flags = ((state->flags & ~ASYNC_INTERNAL_FLAGS) |
		       (info->flags & ASYNC_INTERNAL_FLAGS));
	state->custom_divisor = new_serial.custom_divisor;
	state->type = new_serial.type;
	state->close_delay = new_serial.close_delay * HZ/100;
	state->closing_wait = new_serial.closing_wait * HZ/100;
	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
	info->xmit_fifo_size = state->xmit_fifo_size =
		new_serial.xmit_fifo_size;

	release_region(state->port,8);
	if (change_port || change_irq) {
		/*
		 * We need to shutdown the rckt550 port at the old
		 * port/irq combination.
		 */
		shutdown(info);
		state->irq = new_serial.irq;
		info->port = state->port = new_serial.port;
	}
	if (state->type != PORT_UNKNOWN)
		request_region(state->port,8,"rckt550(set)");

	
check_and_exit:
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "EXIT set_serial_info with struct at %X with type=%d\n", (int)info, state ? state->type : -1);
#endif
	if (!state->port || !state->type)
		return 0;
	if (info->flags & ASYNC_INITIALIZED) {
		if (((old_state.flags & ASYNC_SPD_MASK) !=
		     (state->flags & ASYNC_SPD_MASK)) ||
		    (old_state.custom_divisor != state->custom_divisor)) {
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
				info->tty->alt_speed = 57600;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
				info->tty->alt_speed = 115200;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
				info->tty->alt_speed = 230400;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
				info->tty->alt_speed = 460800;
			change_speed(info, 0);
		}
	} else
		retval = startup(info);
	return retval;
}


/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space.
 */
static int get_lsr_info(struct async_struct * info, unsigned int *value)
{
	unsigned char status;
	unsigned int result;
	unsigned long flags;

	save_flags(flags); cli();
	status = serial_in(info, UART_LSR);
	restore_flags(flags);
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	return put_user(result,value);
}


static int get_modem_info(struct async_struct * info, unsigned int *value)
{
	unsigned char control, status;
	unsigned int result;
	unsigned long flags;

	control = info->MCR;
	save_flags(flags); cli();
	status = serial_in(info, UART_MSR);
	restore_flags(flags);
	result =  ((control & UART_MCR_RTS) ? TIOCM_RTS : 0)
		| ((control & UART_MCR_DTR) ? TIOCM_DTR : 0)
#ifdef TIOCM_OUT1
		| ((control & UART_MCR_OUT1) ? TIOCM_OUT1 : 0)
		| ((control & UART_MCR_OUT2) ? TIOCM_OUT2 : 0)
#endif
		| ((status  & UART_MSR_DCD) ? TIOCM_CAR : 0)
		| ((status  & UART_MSR_RI) ? TIOCM_RNG : 0)
		| ((status  & UART_MSR_DSR) ? TIOCM_DSR : 0)
		| ((status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
	return put_user(result,value);
}

static int set_modem_info(struct async_struct * info, unsigned int cmd,
			  unsigned int *value)
{
	int error;
	unsigned int arg;
	unsigned long flags;

	error = get_user(arg, value);
	if (error)
		return error;
	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR |= UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR |= UART_MCR_OUT2;
#endif
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->MCR &= ~UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR &= ~UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR &= ~UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR &= ~UART_MCR_OUT2;
#endif
		break;
	case TIOCMSET:
		info->MCR = ((info->MCR & ~(UART_MCR_RTS |
#ifdef TIOCM_OUT1
					    UART_MCR_OUT1 |
					    UART_MCR_OUT2 |
#endif
					    UART_MCR_DTR))
			     | ((arg & TIOCM_RTS) ? UART_MCR_RTS : 0)
#ifdef TIOCM_OUT1
			     | ((arg & TIOCM_OUT1) ? UART_MCR_OUT1 : 0)
			     | ((arg & TIOCM_OUT2) ? UART_MCR_OUT2 : 0)
#endif
			     | ((arg & TIOCM_DTR) ? UART_MCR_DTR : 0));
		break;
	default:
		return -EINVAL;
	}
	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
	return 0;
}

static int do_autoconfig(struct async_struct * info)
{
	int			retval;
	
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	
	if (info->state->count > 1)
		return -EBUSY;
	
	shutdown(info);

	autoconfig(info->state);
	if ((info->state->flags & ASYNC_AUTO_IRQ) &&
	    (info->state->port != 0) &&
	    (info->state->type != PORT_UNKNOWN))
		info->state->irq = detect_uart_irq(info->state);

	retval = startup(info);
	if (retval)
		return retval;
	return 0;
}

/*
 * rckt550_break() --- routine which turns the break handling on or off
 */
static void rckt550_break(struct tty_struct *tty, int break_state)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "rckt550_break"))
		return;

	if (!info->port)
		return;
	save_flags(flags); cli();
	if (break_state == -1)
		serial_out(info, UART_LCR,
			   serial_inp(info, UART_LCR) | UART_LCR_SBC);
	else
		serial_out(info, UART_LCR,
			   serial_inp(info, UART_LCR) & ~UART_LCR_SBC);
	restore_flags(flags);
}

static int get_multiport_struct(struct async_struct * info,
				struct serial_multiport_struct *retinfo)
{
	struct serial_multiport_struct ret;
	struct rs_multiport_struct *multi;
	
	multi = &rs_multiport[info->state->irq];

	ret.port_monitor = multi->port_monitor;
	
	ret.port1 = multi->port1;
	ret.mask1 = multi->mask1;
	ret.match1 = multi->match1;
	
	ret.port2 = multi->port2;
	ret.mask2 = multi->mask2;
	ret.match2 = multi->match2;
	
	ret.port3 = multi->port3;
	ret.mask3 = multi->mask3;
	ret.match3 = multi->match3;
	
	ret.port4 = multi->port4;
	ret.mask4 = multi->mask4;
	ret.match4 = multi->match4;

	ret.irq = info->state->irq;

	if (copy_to_user(retinfo,&ret,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_multiport_struct(struct async_struct * info,
				struct serial_multiport_struct *in_multi)
{
	struct serial_multiport_struct new_multi;
	struct rs_multiport_struct *multi;
	struct serial_state *state;
	int	was_multi, now_multi;
	int	retval;
	void (*handler)(int, void *, struct pt_regs *);

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	state = info->state;
	
	if (copy_from_user(&new_multi, in_multi,
			   sizeof(struct serial_multiport_struct)))
		return -EFAULT;
	
	if (new_multi.irq != state->irq || state->irq == 0 ||
	    !IRQ_ports[state->irq])
		return -EINVAL;

	multi = &rs_multiport[state->irq];
	was_multi = (multi->port1 != 0);
	
	multi->port_monitor = new_multi.port_monitor;
	
	if (multi->port1)
		release_region(multi->port1,1);
	multi->port1 = new_multi.port1;
	multi->mask1 = new_multi.mask1;
	multi->match1 = new_multi.match1;
	if (multi->port1)
		request_region(multi->port1,1,"rckt550(multiport1)");

	if (multi->port2)
		release_region(multi->port2,1);
	multi->port2 = new_multi.port2;
	multi->mask2 = new_multi.mask2;
	multi->match2 = new_multi.match2;
	if (multi->port2)
		request_region(multi->port2,1,"rckt550(multiport2)");

	if (multi->port3)
		release_region(multi->port3,1);
	multi->port3 = new_multi.port3;
	multi->mask3 = new_multi.mask3;
	multi->match3 = new_multi.match3;
	if (multi->port3)
		request_region(multi->port3,1,"rckt550(multiport3)");

	if (multi->port4)
		release_region(multi->port4,1);
	multi->port4 = new_multi.port4;
	multi->mask4 = new_multi.mask4;
	multi->match4 = new_multi.match4;
	if (multi->port4)
		request_region(multi->port4,1,"rckt550(multiport4)");

	now_multi = (multi->port1 != 0);
	
	if (IRQ_ports[state->irq]->next_port &&
	    (was_multi != now_multi))
	{
#ifdef SERIAL_DEBUG_IRQ
		printk ( "in set_multiport_struct : Freeing IRQ%d\n", state->irq );
#endif
		free_irq(state->irq, serial_name);
		if (now_multi)
			handler = rs_interrupt_multi;
		else
			handler = rs_interrupt;

#ifdef SERIAL_DEBUG_IRQ
		printk ( "in set_multiport_struct : Requesting IRQ%d\n", state->irq );
#endif
		retval = request_irq(state->irq, handler, SA_SHIRQ,
				     "rckt550", serial_name);
		if (retval) {
			printk("Couldn't reallocate rckt550 interrupt "
			       "driver!!\n");
		}
	}
	return 0;
}

static int rckt550_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	int error;
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct async_icount cprev, cnow;	/* kernel counter temps */
	struct serial_icounter_struct *p_cuser;	/* user space */
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "rckt550_ioctl"))
		return -ENODEV;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}
	
	switch (cmd) {
		case TIOCMGET:
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			return get_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSSERIAL:
			return set_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSERCONFIG:
			return do_autoconfig(info);

		case TIOCSERGETLSR: /* Get line status register */
			return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
			if (copy_to_user((struct async_struct *) arg,
					 info, sizeof(struct async_struct)))
				return -EFAULT;
			return 0;
				
		case TIOCSERGETMULTI:
			return get_multiport_struct(info,
				       (struct serial_multiport_struct *) arg);
		case TIOCSERSETMULTI:
			return set_multiport_struct(info,
				       (struct serial_multiport_struct *) arg);
			
		/*
		 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
		 * - mask passed in arg for lines of interest
 		 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
		 * Caller should use TIOCGICOUNT to see which one it was
		 */
		case TIOCMIWAIT:
			save_flags(flags); cli();
			/* note the counters on entry */
			cprev = info->state->icount;
			restore_flags(flags);
			while (1) {
				interruptible_sleep_on(&info->delta_msr_wait);
				/* see if a signal did it */
				if (signal_pending(current))
					return -ERESTARTSYS;
				save_flags(flags); cli();
				cnow = info->state->icount; /* atomic copy */
				restore_flags(flags);
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
				    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */
				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				     ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			/* NOTREACHED */

		/*
		 * Get counter of input rckt550 line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			save_flags(flags); cli();
			cnow = info->state->icount;
			restore_flags(flags);
			p_cuser = (struct serial_icounter_struct *) arg;
			error = put_user(cnow.cts, &p_cuser->cts);
			if (error) return error;
			error = put_user(cnow.dsr, &p_cuser->dsr);
			if (error) return error;
			error = put_user(cnow.rng, &p_cuser->rng);
			if (error) return error;
			error = put_user(cnow.dcd, &p_cuser->dcd);
			if (error) return error;
			error = put_user(cnow.rx, &p_cuser->rx);
			if (error) return error;
			error = put_user(cnow.tx, &p_cuser->tx);
			if (error) return error;
			error = put_user(cnow.frame, &p_cuser->frame);
			if (error) return error;
			error = put_user(cnow.overrun, &p_cuser->overrun);
			if (error) return error;
			error = put_user(cnow.parity, &p_cuser->parity);
			if (error) return error;
			error = put_user(cnow.brk, &p_cuser->brk);
			if (error) return error;
			error = put_user(cnow.buf_overrun, &p_cuser->buf_overrun);
			if (error) return error;			
			return 0;

		case TIOCSERGWILD:
		case TIOCSERSWILD:
			/* "setserial -W" is called in Debian boot */
			printk ("TIOCSER?WILD ioctl obsolete, ignored.\n");
			return 0;

		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void rckt550_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;
	unsigned int cflag = tty->termios->c_cflag;
	
	if (   (cflag == old_termios->c_cflag)
	    && (   RELEVANT_IFLAG(tty->termios->c_iflag)
		== RELEVANT_IFLAG(old_termios->c_iflag)))
	  return;

	change_speed(info, old_termios);

	/* Handle transition to B0 status */
	if ((old_termios->c_cflag & CBAUD) &&
	    !(cflag & CBAUD)) {
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);
		save_flags(flags); cli();
		serial_out(info, UART_MCR, info->MCR);
		restore_flags(flags);
	}
	
	/* Handle transition away from B0 status */
	if (!(old_termios->c_cflag & CBAUD) &&
	    (cflag & CBAUD)) {
		info->MCR |= UART_MCR_DTR;
		if (!(tty->termios->c_cflag & CRTSCTS) ||
		    !test_bit(TTY_THROTTLED, &tty->flags)) {
			info->MCR |= UART_MCR_RTS;
		}
		save_flags(flags); cli();
		serial_out(info, UART_MCR, info->MCR);
		restore_flags(flags);
	}
	
	/* Handle turning off CRTSCTS */
	if ((old_termios->c_cflag & CRTSCTS) &&
	    !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		rckt550_start(tty);
	}

#if 0
	/*
	 * No need to wake up processes in open wait, since they
	 * sample the CLOCAL flag once, and don't recheck it.
	 * XXX  It's not clear whether the current behavior is correct
	 * or not.  Hence, this may change.....
	 */
	if (!(old_termios->c_cflag & CLOCAL) &&
	    (tty->termios->c_cflag & CLOCAL))
		wake_up_interruptible(&info->open_wait);
#endif
}

/*
 * ------------------------------------------------------------
 * rckt550_close()
 *
 * This routine is called when the rckt550 port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void rckt550_close(struct tty_struct *tty, struct file * filp)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct serial_state *state;
	unsigned long flags;

	if (!info || serial_paranoia_check(info, tty->device, "rckt550_close"))
		return;

	state = info->state;
	
	save_flags(flags); cli();
	
	if (tty_hung_up_p(filp)) {
		DBG_CNT("before DEC-hung");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	
#ifdef SERIAL_DEBUG_OPEN
	printk("rckt550_close ttys%d, count = %d\n", info->line, state->count);
#endif


#if LINUX_VERSION_CODE<VERSION_CODE(2,4,21)
	if ((tty->count == 1) && (state->count != 1))
#else
	if( ( atomic_read(&tty->count) == 1) && (state->count != 1) ) 
#endif
	{
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  state->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * rckt550 port won't be shutdown.
		 */
		printk("rckt550_close: ttys%d: bad rckt550 port count; tty->count is 1, "
		       "state->count is %d\n", info->line, state->count);
		/*
		 * There could still be a problem in this area
		 */
		state->count = 1;
	}
	if (--state->count < 0) {
		printk("rckt550_close: bad rckt550 port count for ttys%d: %d\n",
		       info->line, state->count);
		state->count = 0;
	}
	if (state->count) {
		DBG_CNT("before DEC-2");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	info->flags |= ASYNC_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->state->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->state->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	info->IER &= ~UART_IER_RLSI;
	info->read_status_mask &= ~UART_LSR_DR;
	if (info->flags & ASYNC_INITIALIZED) {
		serial_out(info, UART_IER, info->IER);
		/*
		 * Before we drop DTR, make sure the UART transmitter
		 * has completely drained; this is especially
		 * important if there is a transmit FIFO!
		 */
		rckt550_wait_until_sent(tty, info->timeout);
	}
	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
	restore_flags(flags);
}

/*
 * rckt550_wait_until_sent() --- wait until the transmitter is empty
 */
static void rckt550_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;
	
	if (serial_paranoia_check(info, tty->device, "rckt550_wait_until_sent"))
		return;

	if (info->state->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0)
		return; /* Just in case.... */

	orig_jiffies = jiffies;
	/*
	 * Set the check interval to be 1/5 of the estimated time to
	 * send a single character, and make it at least 1.  The check
	 * interval should also be less than the timeout.
	 *
	 * Note: we have to use pretty tight timings here to satisfy
	 * the NIST-PCTS.
	 */
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;
	if (char_time == 0)
		char_time = 1;
	if (timeout)
	  char_time = MIN(char_time, timeout);
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */
	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("In rckt550_wait_until_sent(%d) check=%lu...", timeout, char_time);
	printk("jiff=%lu...", jiffies);
#endif
	while (!((lsr = serial_inp(info, UART_LSR)) & UART_LSR_TEMT)) {
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
		printk("lsr = %d (jiff=%lu)...", lsr, jiffies);
#endif
		current->state = TASK_INTERRUPTIBLE;

		/* make us low-priority */
		#if LINUX_VERSION_CODE<VERSION_CODE(2,4,18)
			current->counter = 0;	
		#endif
		
		schedule_timeout(char_time);
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
	current->state = TASK_RUNNING;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	printk("lsr = %d (jiff=%lu)...done\n", lsr, jiffies);
#endif
}

/*
 * rckt550_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void rckt550_hangup(struct tty_struct *tty)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct serial_state *state = info->state;
	
	if (serial_paranoia_check(info, tty->device, "rckt550_hangup"))
		return;

	state = info->state;
	
	rckt550_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	state->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 * ------------------------------------------------------------
 * rckt550_open() and friends
 * ------------------------------------------------------------
 */
static int SetPortMode ( struct serial_state *state, int LineMode )
{
	int i;
   int RetVal = FALSE;

#ifdef SERIAL_DEBUG_OPEN
	printk ( "SetPortMode %x (base=%x) for %x\n", state, state->port, LineMode );
#endif

	for ( i=0 ; i<NumberOfDetectedBoards ; i++ )
	{
#ifdef SERIAL_DEBUG_OPEN
		printk ( "Is it board %d, at %x\n", i, BoardData[i].BoardBase );
#endif
      if ( ( state->port & 0xff00 ) == BoardData[i].BoardBase )
		{
#ifdef SERIAL_DEBUG_OPEN
			printk ( "Configbase=%x\n", BoardData[i].RsModeConfigBase );
#endif
			if ( BoardData[i].RsModeConfigBase )                   /* This is zero for non-soft boards */
			{
            int RsModeConfigPort = BoardData[i].RsModeConfigBase + ( ( state->port & 0xff ) >> 3 );

#ifdef SERIAL_DEBUG_OPEN
				printk ( "Setting port %x to %x\n",RsModeConfigPort,LineMode); 
#endif
            outb(LineMode,RsModeConfigPort);
				if ( LineMode == ROCKET550_RS485 )
				{
					WriteACR(state,DTR_ENABLE_ACTIVE_LOW,~(OXSEMI_ACR_3|OXSEMI_ACR_4));
				}
				RetVal = TRUE;                                 /* Show that we did it */
			}
			break;
		}
	}

	return RetVal;                                         /* TRUE=OK, FALSE=not soft board */
}

static unsigned char ReadACR ( struct serial_state *state )
{
	unsigned char Acr;
	unsigned char Lcr = inb(state->port+UART_LCR);

	outb(0,state->port+UART_LCR);                               /* not 0xBF */
	
	outb ( 0, state->port+UART_SPR);							/* 0 selects ACR */
	Acr = inb ( state->port+UART_ICR );                         /* get old value */

#ifdef SERIAL_DEBUG_OPEN
   printk ( "Read ACR as %02x\n", Acr );
#endif

	outb ( Lcr, state->port+UART_LCR );                         /* restore LCR */

	return Acr;
}

static void WriteACR ( struct serial_state *state, unsigned char OrBits, unsigned char AndBits )
{
   unsigned char Val = ReadACR(state);
	unsigned char Lcr = inb(state->port+UART_LCR);

	outb(0,state->port+UART_LCR);                               /* not 0xBF */

	Val &= AndBits;
	Val |= OrBits;

	outb ( 0, state->port+UART_SPR);										/* 0 selects ACR */
	outb ( Val, state->port+UART_ICR);

#ifdef SERIAL_DEBUG_OPEN
   printk ( "Wrote ACR as %02x\n", Val );
#endif

	outb ( Lcr, state->port+UART_LCR );                         /* restore LCR */
}

/* -------------------------------------------------------------*/

/****************************************************************
	WriteTTL
	
	Write the transmit FIFO trigger level to OX16C954
	UART TTL register.

	Assumes that 0X16C954 has been configured for 
	enhanced mode operation (EFR4:1)
	and ACR5:=1 (trigger level enable)
*****************************************************************/
/*
static void WriteTTL ( struct serial_state *state, unsigned char val )
{
	outb( 0x04, state->port+UART_SPR );
	outb( val, state->port+UART_ICR );
}
*/
/****************************************************************
	WriteRTL
	
	Write the receive FIFO trigger level to OX16C954
	UART RTL register.

	Assumes that 0X16C954 has been configured for 
	enhanced mode operation (EFR4:1)
	and ACR5:=1 (trigger level enable)
*****************************************************************/
/*
static void WriteRTL ( struct serial_state *state, unsigned char val )
{
	
	
	outb( 0x05, state->port+UART_SPR );
	outb( val, state->port+UART_ICR );
}
*/

static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct async_struct *info)
{
#if	EGO_FIX_1
	DECLARE_WAITQUEUE (wait, current);
#else
	struct wait_queue wait = { current, NULL };
#endif
	struct serial_state *state = info->state;
	int		retval;
	int		do_clocal = 0, extra_count = 0;
	unsigned long	flags;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) ||
	    (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

#ifdef SERIAL_DEBUG_OPEN
	printk ( "Driver Subtype=%04x\n", tty->driver.subtype );
#endif


	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT)
	{
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		SetPortMode ( state, ROCKET550_RS232 );
		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}
	
	/*
	 * If this is a RS422 device, then make sure the normal
	 * or callout device isn't being used.
	 */
	if ( strncmp(tty->driver.name,"Rckt550-422",sizeof("Rckt550-422") ) == 0 )
	{
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		if ( !SetPortMode ( state, ROCKET550_RS422 ) )
		{
				  return -EINVAL;                        /* not a 422-capable device */
		}
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}
	
	/*
	 * If this is a RS485 device, then make sure the normal
	 * or callout device isn't being used.
	 */
	if ( strncmp(tty->driver.name,"Rckt550-485",sizeof("Rckt550-485") ) == 0 )
	{
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
		    return -EBUSY;
		if ( !SetPortMode ( state, ROCKET550_RS485 ) )
		{
		   return -EINVAL;                        /* not a 485-capable device */
	   }
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}
	
   SetPortMode ( state, ROCKET550_RS232 );

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & ASYNC_CALLOUT_ACTIVE) {
		if (state->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, state->count is dropped by one, so that
	 * rckt550_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready before block: ttys%d, count = %d\n",
	       state->line, state->count);
#endif
	save_flags(flags); cli();
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		state->count--;
	}
	restore_flags(flags);
	info->blocked_open++;
	while (1) {
		save_flags(flags); cli();
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (tty->termios->c_cflag & CBAUD))
			serial_out(info, UART_MCR,
				   serial_inp(info, UART_MCR) |
				   (UART_MCR_DTR | UART_MCR_RTS));
		restore_flags(flags);
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
		    !(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
#else
			retval = -EAGAIN;
#endif
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready unblocking: hung up\n" );
#endif
			break;
		}
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    !(info->flags & ASYNC_CLOSING) &&
		    (do_clocal || (serial_in(info, UART_MSR) &
				   UART_MSR_DCD)))
		{
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready unblocking: MSR %x\n", serial_in(info,UART_MSR) );
#endif
			break;
		}
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready unblocking: signal\n" );
#endif
			break;
		}
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready blocking: ttys%d, count = %d\n",
		       info->line, state->count);
#endif
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (extra_count)
		state->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready after blocking: ttys%d, count = %d\n",
	       info->line, state->count);
#endif
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
}

static int get_async_struct(int line, struct async_struct **ret_info)
{
	struct async_struct *info;
	struct serial_state *sstate;

	sstate = rckt550_table + line;
	sstate->count++;
	if (sstate->info) {
		*ret_info = sstate->info;
		return 0;
	}
	info = kmalloc(sizeof(struct async_struct), GFP_KERNEL);
	if (!info) {
		sstate->count--;
		return -ENOMEM;
	}
	memset(info, 0, sizeof(struct async_struct));
	info->magic = SERIAL_MAGIC;
#ifdef	EGO_FIX_1
	init_waitqueue_head(&info->open_wait);
	init_waitqueue_head(&info->close_wait);
	init_waitqueue_head(&info->delta_msr_wait);
#endif
	info->port = sstate->port;
	info->flags = sstate->flags;
	info->xmit_fifo_size = sstate->xmit_fifo_size;
	info->line = line;
	info->tqueue.routine = do_softint;
	info->tqueue.data = info;
	info->state = sstate;
	if (sstate->info) {
#if EGO_FIX_1
		kfree(info);
#else
		kfree_s(info, sizeof(struct async_struct));
#endif
		*ret_info = sstate->info;
		return 0;
	}
	*ret_info = sstate->info = info;
	return 0;
}

/*
 * This routine is called whenever a rckt550 port is opened.  It
 * enables interrupts for a rckt550 port, linking in its async structure into
 * the IRQ chain.   It also performs the rckt550-specific
 * initialization for the tty structure.
 */
static int rckt550_open(struct tty_struct *tty, struct file * filp)
{
	struct async_struct	*info;
	int 			retval, line;
	unsigned long		page;
	
	MOD_INC_USE_COUNT;
	line = MINOR(tty->device) - tty->driver.minor_start;
	
	if ((line < 0) || (line >= NR_PORTS)) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}

	tprint ( "RCKT_OPEN: Minor=%d\n", MINOR(tty->device) );

	if ( rckt550_table[MINOR(tty->device)].port == 0) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}

	retval = get_async_struct(line, &info);
	if (retval) {
		MOD_DEC_USE_COUNT;
		return retval;
	}

	info->state->baud_base = 115200 * 4;  // Kludge for 7.2Mhz clocked Rockets
	tprint ( "RCKT_OPEN:line %d, baud_base=%d\n", line, info->state->baud_base );

	tty->driver_data = info;
	info->tty = tty;
	if (serial_paranoia_check(info, tty->device, "rckt550_open")) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this */
		return -ENODEV;
	}

#ifdef SERIAL_DEBUG_OPEN
	printk("rckt550_open %s%d, count = %d\n", tty->driver.name, info->line,
	       info->state->count);
#endif
	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;

	if (!tmp_buf) {
		page = get_free_page(GFP_KERNEL);
		if (!page) {
			/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
			return -ENOMEM;
		}
		if (tmp_buf)
			free_page(page);
		else
			tmp_buf = (unsigned char *) page;
	}

	/*
	 * If the port is the middle of closing, bail out now
	 */
	if (tty_hung_up_p(filp) ||
	    (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

	/*
	 * Start up rckt550 port
	 */
	retval = startup(info);
	if (retval) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
		return retval;
	}

	retval = block_til_ready(tty, filp, info);
	if (retval) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
#ifdef SERIAL_DEBUG_OPEN
		printk("rckt550_open returning after block_til_ready with %d\n",
		       retval);
#endif
		return retval;
	}

	if ((info->state->count == 1) &&
	    (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->state->normal_termios;
		else
			*tty->termios = info->state->callout_termios;
		change_speed(info, 0);
	}
#ifdef CONFIG_SERIAL_CONSOLE
	if (sercons.cflag && sercons.index == line) {
		tty->termios->c_cflag = sercons.cflag;
		sercons.cflag = 0;
		change_speed(info, 0);
	}
#endif
	info->session = current->session;
	info->pgrp = current->pgrp;

#ifdef SERIAL_DEBUG_OPEN
	printk("rckt550_open ttys%d successful...", info->line);
#endif
	return 0;
}

/*
 * /proc fs routines....
 */

static inline int line_info(char *buf, struct serial_state *state)
{
	struct async_struct *info = state->info, scr_info;
	char	stat_buf[30], control, status;
	int	ret;
	unsigned long flags;

#if 0 /* EGO_FIX_1 */
	printk(KERN_INFO "EGO: rckt550: line: %d: uart:%s (type %d) port:%lX irq:%d\n",
		      state->line, uart_config[state->type].name, state->type,
		      state->port, state->irq);
	ret = sprintf(buf, "%d: uart:%s (type no %d) port:%X irq:%d",
		      state->line, uart_config[state->type].name, state->type,
		      state->port, state->irq);
#else
	ret = sprintf(buf, "%d: uart:%s port:%X irq:%d",
		      state->line, uart_config[state->type].name,
		      (unsigned int)state->port, state->irq);
#endif

	if (!state->port || (state->type == PORT_UNKNOWN)) {
		ret += sprintf(buf+ret, "\n");
		return ret;
	}

	/*
	 * Figure out the current RS-232 lines
	 */
	if (!info) {
		info = &scr_info;	/* This is just for serial_{in,out} */

		info->magic = SERIAL_MAGIC;
		info->port = state->port;
		info->flags = state->flags;
		info->quot = 0;
		info->tty = 0;
	}
	save_flags(flags); cli();
	status = serial_in(info, UART_MSR);
	control = info ? info->MCR : serial_in(info, UART_MCR);
	restore_flags(flags);
	
	stat_buf[0] = 0;
	stat_buf[1] = 0;
	if (control & UART_MCR_RTS)
		strcat(stat_buf, "|RTS");
	if (status & UART_MSR_CTS)
		strcat(stat_buf, "|CTS");
	if (control & UART_MCR_DTR)
		strcat(stat_buf, "|DTR");
	if (status & UART_MSR_DSR)
		strcat(stat_buf, "|DSR");
	if (status & UART_MSR_DCD)
		strcat(stat_buf, "|CD");
	if (status & UART_MSR_RI)
		strcat(stat_buf, "|RI");

	if (info->quot) {
		ret += sprintf(buf+ret, " baud:%d",
			       state->baud_base / info->quot);
	}

	ret += sprintf(buf+ret, " tx:%d rx:%d",
		      state->icount.tx, state->icount.rx);

	if (state->icount.frame)
		ret += sprintf(buf+ret, " fe:%d", state->icount.frame);
	
	if (state->icount.parity)
		ret += sprintf(buf+ret, " pe:%d", state->icount.parity);
	
	if (state->icount.brk)
		ret += sprintf(buf+ret, " brk:%d", state->icount.brk);	

	if (state->icount.overrun)
		ret += sprintf(buf+ret, " oe:%d", state->icount.overrun);

	/*
	 * Last thing is the RS-232 status lines
	 */
	ret += sprintf(buf+ret, " %s\n", stat_buf+1);
	return ret;
}

int rckt550_read_proc(char *page, char **start, off_t idx, ssize_t count,
		 int *eof, void *data)
{
	int n;

#if RCKT_DEBUG_MORE
	printk("EGO read_proc called with page=%X, start=%X, *start=%X, idx[%d]=%d, count=%d, *eof=%d, data=%X\n",
			(int) page, (int) start, (int) *start, sizeof (off_t), (int) idx, count, *eof, (int) data);
#endif

	*start = (char *) 1L;

	if (!idx) {
#ifdef	EGO_FIX_1
	{
	char temp[4096];	/* More than ample */

	n = sprintf(temp, "serinfo:1.0 driver:%s\n", serial_version);
	if (n < count && n > 0)
	{
		memcpy (page, temp, n);
		*(page + n) = '\0';	/* And clear the line terminator */
	}
	}
#else
		n = sprintf(page, "serinfo:1.0 driver:%s\n", serial_version);
#endif
		goto out;
	}
	idx--;

	n = 0;
	if (idx >= NR_PORTS || idx < 0)
		goto out;

#ifdef	EGO_FIX_1
	{
	char temp[4096];	/* More than ample */

	n = line_info(temp, &rckt550_table[idx]);	/* saves carrying count into line_info */
	if (n < count && n > 0)
	{
		memcpy (page, temp, n);
		*(page + n) = '\0';	/* And clear the line terminator */
	}
	}
#else
	n = line_info(page, &rckt550_table[idx]);
#endif
	if (idx + 1 == NR_PORTS)
		*eof = 1;

 out:
#ifdef	EGO_FIX_1
	if (n >= count || n < 0)
#else
	if (n > count)
#endif
		n = 0;
	return n;
}

/*
 * ---------------------------------------------------------------------
 * rs_init() and friends
 *
 * rs_init() is called at boot-time to initialize the rckt550 driver.
 * ---------------------------------------------------------------------
 */

/*
 * This routine prints out the appropriate rckt550 driver version
 * number, and identifies which options were configured into this
 * driver.
 */
static _INLINE_ void show_serial_version(void)
{
#ifdef	EGO_FIX_1
 	printk(KERN_INFO "%s version %s (with EgO Fixes of 7 AUG 2001 & 8 Aug 2002) with", serial_name, serial_version);
#else
 	printk(KERN_INFO "%s version %s with", serial_name, serial_version);
#endif
	printk(" MANY_PORTS");
	printk(" MULTIPORT");
	printk(" SHARE_IRQ");
	printk(" enabled\n");
}

/*
 * This routine detect the IRQ of a rckt550 port by clearing OUT2 when
 * no UART interrupt are requested (IER = 0) (*GPL*). This seems to work at
 * each time, as long as no other device permanently request the IRQ.
 * If no IRQ is detected, or multiple IRQ appear, this function returns 0.
 * The variable "state" and the field "state->port" should not be null.
 */
static unsigned detect_uart_irq (struct serial_state * state)
{
	int irq;
	unsigned long irqs;
	unsigned char save_mcr, save_ier;
	struct async_struct scr_info; 
	unsigned char save_ICP=0; /* no warning */
	unsigned short ICP=0;

	if (state->flags & ASYNC_FOURPORT)  {
		ICP = (state->port & 0xFE0) | 0x01F;
		save_ICP = inb_p(ICP);
		outb_p(0x80, ICP);
		(void) inb_p(ICP);
	}

	scr_info.magic = SERIAL_MAGIC;
	scr_info.port = state->port;
	scr_info.flags = state->flags;

	/* forget possible initially masked and pending IRQ */
	probe_irq_off(probe_irq_on());
	save_mcr = serial_inp(&scr_info, UART_MCR);
	save_ier = serial_inp(&scr_info, UART_IER);
	serial_outp(&scr_info, UART_MCR, UART_MCR_OUT1 | UART_MCR_OUT2);
	
	irqs = probe_irq_on();
	serial_outp(&scr_info, UART_MCR, 0);
	udelay (10);
	if (state->flags & ASYNC_FOURPORT)  {
		serial_outp(&scr_info, UART_MCR,
			    UART_MCR_DTR | UART_MCR_RTS);
	} else {
		serial_outp(&scr_info, UART_MCR,
			    UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2);
	}
	serial_outp(&scr_info, UART_IER, 0x0f);	/* enable all intrs */
	(void)serial_inp(&scr_info, UART_LSR);
	(void)serial_inp(&scr_info, UART_RX);
	(void)serial_inp(&scr_info, UART_IIR);
	(void)serial_inp(&scr_info, UART_MSR);
	serial_outp(&scr_info, UART_TX, 0xFF);
	udelay (20);
	irq = probe_irq_off(irqs);

	serial_outp(&scr_info, UART_MCR, save_mcr);
	serial_outp(&scr_info, UART_IER, save_ier);
	if (state->flags & ASYNC_FOURPORT)
		outb_p(save_ICP, ICP);
	return (irq > 0)? irq : 0;
}

/*
 * This routine is called by rs_init() to initialize a specific rckt550
 * port.  
 */

static void autoconfig(struct serial_state * state)
{
	unsigned char status1, status2, scratch, scratch2,efrval;
	struct async_struct *info, scr_info;
	unsigned long flags;
	
	if (!state->port)
		return;
		
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "ON ENTRY to autoconfig %X has type=%d\n", state, state->type);
#endif
	info = &scr_info;	/* This is just for serial_{in,out} */

	info->magic = SERIAL_MAGIC;
	info->port = state->port;
	info->flags = state->flags;

	save_flags(flags); cli();
	
	/*
	 * Do a simple existence test first; if we fail this, there's
	 * no point trying anything else.
	 *
	 * 0x80 is used as a nonsense port to prevent against FALSE
	 * positives due to ISA bus float.  The assumption is that
	 * 0x80 is a non-existent port; which should be safe since
	 * include/asm/io.h also makes this assumption.
	 */
	scratch = serial_inp(info, UART_IER);
	serial_outp(info, UART_IER, 0);
	outb(0xff, 0x080);
	scratch2 = serial_inp(info, UART_IER);
	serial_outp(info, UART_IER, scratch);
	if (scratch2) {
		restore_flags(flags);
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "EGO (B) Wrote 0 Read =%X\n", scratch2);
#endif
		return;		/* We failed; there's nothing here */
	}

	/*
	 * Check to see if a UART is really there.  Certain broken
	 * internal modems based on the Rockwell chipset fail this
	 * test, because they apparently don't implement the loopback
	 * test mode.  So this test is skipped on the COM 1 through
	 * COM 4 ports.  This *should* be safe, since no board
	 * manufacturer would be stupid enough to design a board
	 * that conflicts with COM 1-4 --- we hope!
	 */
	if (!(state->flags & ASYNC_SKIP_TEST)) {
		scratch = serial_inp(info, UART_MCR);
		serial_outp(info, UART_MCR, UART_MCR_LOOP | scratch);
		serial_outp(info, UART_MCR, UART_MCR_LOOP | 0x0A);
		status1 = serial_inp(info, UART_MSR) & 0xF0;
		serial_outp(info, UART_MCR, scratch);
		if (status1 != 0x90) {
			restore_flags(flags);
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "EGO (C) state->type =%d\n", state->type);
#endif
			return;
		}
	}
	
	scratch2 = serial_in(info, UART_LCR);
	serial_outp(info, UART_LCR, 0xBF); /* set up for StarTech test */
	serial_outp(info, UART_EFR, 0);	/* EFR is the same as FCR */
	serial_outp(info, UART_LCR, 0);
	serial_outp(info, UART_FCR, UART_FCR_ENABLE_FIFO);
	scratch = serial_in(info, UART_IIR) >> 6;
	switch (scratch) {
		case 0:
			state->type = PORT_16450;
			break;
		case 1:
			state->type = PORT_UNKNOWN;
			break;
		case 2:
			state->type = PORT_16550;
			break;
		case 3:
			state->type = PORT_16550A;
			break;
	}
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "EGO (A) state->type =%d\n", state->type);
#endif
	if (state->type == PORT_16550A) {
		/* Check for Startech UART's */
		serial_outp(info, UART_LCR, scratch2 | UART_LCR_DLAB);
		if (serial_in(info, UART_EFR) == 0) {
			state->type = PORT_16650;
		} else {
			serial_outp(info, UART_LCR, 0xBF);
			if (serial_in(info, UART_EFR) == 0)
			{
				state->type = PORT_16650V2;	/* As on the OX16C954 */
				
				/*enable 16650 mode REL2.03 modification for DCS*/
				efrval = serial_in(info, UART_EFR ) | UART_EFR_ECB;	/*switch on enhanced mode*/
				serial_outp(info,UART_EFR,efrval);

#if 0 /*RCKT_DEBUG_MORE */
				/*
				 * EGO put it in to try to see if it really was
				 * an OX16C954 chip that was connected.
				 * This is taken out as it does not seem to work
				 * but doesnot seen to be required anyway
				 */
				printk (KERN_INFO "EGO: Have 16650V2 - see if OX16C954\n");
				{
				/*
				 * Now see if it is an oxford chip
				 */
					unsigned int tempdata[4];
					unsigned char ACR_temp = 0;

					scratch = serial_in(info, UART_LCR);	/* Read current value */
					serial_outp(info, UART_LCR, 0x0);	/* Write (any but 0xBF) to Line Control Register */

					serial_outp(info, UART_SCR, 0x0);	/* UART_SPR: 0 will select ACR */

					ACR_temp = (unsigned char) (info->ACR | (1 << 6)); /* Set ICR read enable */
					
					serial_outp(info, UART_LSR, ACR_temp);	/* Set ICR (bit 6) in ACR */
					serial_outp(info, UART_SCR, 0x8);	/* set index address */
					tempdata[0] = serial_in(info, UART_LSR);	/* Read 8 */
					serial_outp(info, UART_SCR, 0x9);	/* set index address */
					tempdata[1] = serial_in(info, UART_LSR);	/* Read 9 */
					serial_outp(info, UART_SCR, 0xA);	/* set index address */
					tempdata[2] = serial_in(info, UART_LSR);	/* Read A */
					serial_outp(info, UART_SCR, 0xB);	/* set index address */
					tempdata[3] = serial_in(info, UART_LSR);	/* Read B */
					serial_outp(info, UART_SCR, 0x0);	/* set index address to ACR again */
					ACR_temp |= (~(1 << 6));
					serial_outp(info, UART_LSR, ACR_temp);	/* Clear ICR (bit 6) in ACR */
					serial_outp(info, UART_LCR, scratch);	/* Reset LCR */
			printk (KERN_INFO "EGO: OCTART VERSION =%X,%X,%X,%X, LSR=%X,SCR=%X, base=%X)\n", tempdata[0],tempdata[1],tempdata[2],tempdata[3], UART_LSR, UART_SCR, (int) info->port);
				}
#endif
			}
		}
	}
	if (state->type == PORT_16550A)
	{
		/* Check for TI 16750 */
		serial_outp(info, UART_LCR, scratch2 | UART_LCR_DLAB);
		serial_outp(info, UART_FCR,
			    UART_FCR_ENABLE_FIFO | UART_FCR7_64BYTE);
		scratch = serial_in(info, UART_IIR) >> 5;
		if (scratch == 7) 
		{
			serial_outp(info, UART_LCR, 0);
			scratch = serial_in(info, UART_IIR) >> 5;
			if (scratch == 6)
				state->type = PORT_16750;
		}
		serial_outp(info, UART_FCR, UART_FCR_ENABLE_FIFO);
	}

	serial_outp(info, UART_LCR, scratch2);	/*this ensures that LCR != 0xBF*/

	if (state->type == PORT_16650V2)
	{
		/*
		  REL2.03 modification for DCS
		  Enhanced mode (EFR4:=1)already enabled. 
		  Enable RTL and TTL trigger level registers to override FCR4 & FCR5
		  Set RX Fifo default trigger level (ICR - RTL register)
		  Set TX Fifo default trigger level (ICR - TTL register)
		*/
		
		serial_outp( info, UART_SPR, 0 );	/*Select ACR register as destination for next ICR write*/
		serial_outp( info, UART_ICR, 0x20); /*enable enhanced mode fifo trigger levels*/
		serial_outp( info, UART_SPR, 4 );	/*Select TX Fifo trigger level register*/
		serial_outp( info, UART_ICR, 1 );   /*Initialise to 1*/
		serial_outp( info, UART_SPR, 5 );	/*Select RX Fifo trigger level register*/
		serial_outp( info, UART_ICR, 8 );	/*Initialise to 8*/
	}
	
	if (state->type == PORT_16450) 
	{
		scratch = serial_in(info, UART_SCR);
		serial_outp(info, UART_SCR, 0xa5);
		status1 = serial_in(info, UART_SCR);
		serial_outp(info, UART_SCR, 0x5a);
		status2 = serial_in(info, UART_SCR);
		serial_outp(info, UART_SCR, scratch);

		if ((status1 != 0xa5) || (status2 != 0x5a))
		{
			state->type = PORT_8250;
		}
	}
	state->xmit_fifo_size =	uart_config[state->type].dfl_xmit_fifo_size;

	if (state->type == PORT_UNKNOWN) {
		restore_flags(flags);
		return;
	}

	request_region(info->port,8,"rckt550(auto)");

	/*
	 * Reset the UART.
	 */
#if defined(__alpha__) && !defined(CONFIG_PCI)
	/*
	 * I wonder what DEC did to the OUT1 and OUT2 lines?
	 * clearing them results in endless interrupts.
	 */
	serial_outp(info, UART_MCR, 0x0c);
#else
	serial_outp(info, UART_MCR, 0x00);
#endif
	serial_outp(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
				     UART_FCR_CLEAR_XMIT));
	(void)serial_in(info, UART_RX);
	serial_outp(info, UART_IER, 0);
	
	restore_flags(flags);
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "ON EXIT from autoconfig %X has type=%d(%s)\n", state, state->type, uart_config[state->type].name);
#endif
}

int register_serial(struct serial_struct *req);
void unregister_serial(int line);

EXPORT_SYMBOL(register_serial);
EXPORT_SYMBOL(unregister_serial);

void CreateBoard ( unsigned long first_chip_base, unsigned long second_chip_base, 
					    int Irq, int PortCount, int Universal, int SoftCapable )
{
	int i;
	
	//printk(KERN_INFO "enter CreateBoard\n");
	BoardData[NumberOfDetectedBoards].NumberOfPortsOnBoard = PortCount;
   BoardData[NumberOfDetectedBoards].BoardBase            = first_chip_base;
	BoardData[NumberOfDetectedBoards].RsModeConfigBase     = 0;

	printk(KERN_INFO "Port Count=%d, Base Addr=%x\n",PortCount,(int)first_chip_base);
			
	if ( Universal )
	{
	   for ( i=0 ; i<PortCount ; i++ )
	   {
		   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].port = first_chip_base + 8 * i;
		   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].irq  = Irq;
	   }
		if ( SoftCapable )
		{
			switch ( PortCount )
			{
	         case 4:
			   {
				   BoardData[NumberOfDetectedBoards].RsModeConfigBase = first_chip_base + 0x20;
					break;
				}
	         case 8:
			   {
				   BoardData[NumberOfDetectedBoards].RsModeConfigBase = first_chip_base + 0x40;
					break;
				}
	         case 16:
			   {
				   BoardData[NumberOfDetectedBoards].RsModeConfigBase = first_chip_base + 0x80;
					break;
				}
			}
		}					  
	}
	else
	{
	   for ( i=0 ; i<OX954_PORTS_PER_CHIP ; i++ )
	   {
		   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].port = first_chip_base + 8 * i;
		   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].irq  = Irq;
	   }
	
	   if ( second_chip_base )
	   {
		   for ( ; i<PortCount ; i++ )
		   {
			   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].port = second_chip_base + 8 * (i-OX954_PORTS_PER_CHIP);
			   rckt550_table[i+NumberOfDetectedBoards*MAX_PORTS_PER_BOARD].irq  = Irq;
		   }
	   }
   }

   NumberOfDetectedBoards++;
}

static inline int DetectComtrolPciBoards ( void )
{
   struct pci_dev *dev_A = NULL;
   struct pci_dev *dev_B = NULL;
   unsigned long first_chip_base, second_chip_base;
   RocketPortDefinition *BoardDef = BoardDefinitions;
   int    SoftCapableBoardDetected = FALSE;

   NumberOfDetectedBoards = 0;

   while ( BoardDef->VendorId )
	{
#ifdef DEBUG_PCI
	   printk ( "RCKT550: Looking for id=%x (%s) boards\n", BoardDef->DeviceId, BoardDef->Description );
#endif
	   while ( ( dev_A = pci_find_device(BoardDef->VendorId, BoardDef->DeviceId, dev_A) ) )
	   {
		   first_chip_base  = PCI_BASE_ADDRESS(dev_A, BoardDef->PortBarIndex);
         second_chip_base = 0;
#ifdef DEBUG_PCI
		   printk ( "%s %d: Base=%lx, Irq=%d\n", BoardDef->Description, NumberOfDetectedBoards, first_chip_base, dev_A->irq );
#endif
	   	if ( BoardDef->SubDeviceId )
			{
            dev_B = NULL;
				while ( ( dev_B = pci_find_device(BoardDef->VendorId, BoardDef->SubDeviceId, dev_B) ) )
				{
			      if ( ( dev_A->bus == dev_B->bus ) &&
                    ( PCI_SLOT(dev_A->devfn) == PCI_SLOT(dev_B->devfn) ) )
					{
  			         second_chip_base = PCI_BASE_ADDRESS(dev_B, 0);                       // BarIndex always zero for Sub-devices
#ifdef DEBUG_PCI
			         printk ( "Rckt550: Subdevice %x for device %x found\n", BoardDef->SubDeviceId, BoardDef->DeviceId );
	  		         printk ( "         %s, Base=%lx, Irq=%d\n", BoardDef->Description, second_chip_base, dev_B->irq );
#endif
					}
				}

				if ( second_chip_base == 0 )
				{
			      printk ( "Rckt550: Subdevice %x for device %x not found\n", BoardDef->SubDeviceId, BoardDef->DeviceId );
               continue;
				}
		   }
		   printk ( "RCKT550:%s %d: Base=%lx, Irq=%d\n", BoardDef->Description, NumberOfDetectedBoards, first_chip_base, dev_A->irq );
			CreateBoard ( first_chip_base, second_chip_base, dev_A->irq, BoardDef->PortCount, BoardDef->Universal, BoardDef->SoftCapable );
			if ( BoardDef->SoftCapable )
			{
            SoftCapableBoardDetected = TRUE;
			}
		}
		
		BoardDef++;
	}

	return SoftCapableBoardDetected;
}


/*
 * The rckt550 driver boot-time initialization code!
 */
#if EGO_FIX_1
__init int rckt550_init(void)
#else
__initfunc(int rckt550_init(void))
#endif
{
	int i;
	struct serial_state * state;
	extern void atomwide_serial_init (void);
	extern void dualsp_serial_init (void);
	int CalloutMajor;
	

#ifdef CONFIG_ATOMWIDE_SERIAL
	atomwide_serial_init ();
#endif
#ifdef CONFIG_DUALSP_SERIAL
	dualsp_serial_init ();
#endif

	init_bh(SERIAL_BH, do_serial_bh);

	for (i = 0; i < NR_IRQS; i++) {
		IRQ_ports[i] = 0;
		IRQ_timeout[i] = 0;
		memset(&rs_multiport[i], 0,
		       sizeof(struct rs_multiport_struct));
	}
#ifdef CONFIG_SERIAL_CONSOLE
	/*
	 *	The interrupt of the rckt550 console port
	 *	can't be shared.
	 */
	if (sercons.flags & CON_CONSDEV) {
		for(i = 0; i < NR_PORTS; i++)
			if (i != sercons.index &&
			    rckt550_table[i].irq == rckt550_table[sercons.index].irq)
				rckt550_table[i].irq = 0;
	}
#endif
	show_serial_version();

	memset ( &serial_driver,  0, sizeof(struct tty_driver) );
	memset ( &callout_driver, 0, sizeof(struct tty_driver) );
	memset ( &RS422_driver,   0, sizeof(struct tty_driver) );
	memset ( &RS485_driver,   0, sizeof(struct tty_driver) );
	
	/* find some Major device numbers we can use */
		
	serial_driver.major = register_chrdev(0,"Rckt550d",(struct file_operations *)-1);
	if ( serial_driver.major == 0 )
	{
		return -ENODEV;
	}
	
	CalloutMajor = register_chrdev(0,"Rckt550m",(struct file_operations *)-1);
	if ( CalloutMajor == 0 )
	{
		unregister_chrdev(serial_driver.major,"Rckt550d" );
		return -ENODEV;
	}
	
	printk ( "rckt550 V%s: Direct MAJOR is %d, callout is %d\n", serial_version,serial_driver.major, CalloutMajor );
	
	unregister_chrdev(serial_driver.major,"Rckt550d" );
	unregister_chrdev(CalloutMajor,"Rckt550m" );

	/* Initialize the tty_driver structure */
	
	serial_driver.magic 					= TTY_DRIVER_MAGIC;
	serial_driver.driver_name 			= "Rckt550";
	serial_driver.name 					= "Rckt550D";
	serial_driver.minor_start 			= 0;
	serial_driver.num 					= NR_PORTS;
	serial_driver.type 					= TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype 				= SERIAL_TYPE_NORMAL;
	serial_driver.init_termios 		= tty_std_termios;
	serial_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags 					= TTY_DRIVER_REAL_RAW;
	serial_driver.refcount 				= &serial_refcount;
	serial_driver.table 					= serial_table;
	serial_driver.termios 				= serial_termios;
	serial_driver.termios_locked 		= serial_termios_locked;

	serial_driver.open 					= rckt550_open;
	serial_driver.close 					= rckt550_close;
	serial_driver.write 					= rckt550_write;
	serial_driver.put_char 				= rckt550_put_char;
	serial_driver.flush_chars 			= rckt550_flush_chars;
	serial_driver.write_room 			= rckt550_write_room;
	serial_driver.chars_in_buffer 	= rckt550_chars_in_buffer;
	serial_driver.flush_buffer 		= rckt550_flush_buffer;
	serial_driver.ioctl 					= rckt550_ioctl;
	serial_driver.throttle 				= rckt550_throttle;
	serial_driver.unthrottle 			= rckt550_unthrottle;
	serial_driver.send_xchar 			= rckt550_send_xchar;
	serial_driver.set_termios 			= rckt550_set_termios;
	serial_driver.stop 					= rckt550_stop;
	serial_driver.start 					= rckt550_start;
	serial_driver.hangup 				= rckt550_hangup;
	serial_driver.break_ctl 			= rckt550_break;
	serial_driver.wait_until_sent 	= rckt550_wait_until_sent;
	serial_driver.read_proc 			= rckt550_read_proc;
	
	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver 			   = serial_driver;
	callout_driver.name  	   = "Rckt550M";
	callout_driver.major			= CalloutMajor;
	callout_driver.subtype 		= SERIAL_TYPE_CALLOUT;
	callout_driver.read_proc 	= 0;
	callout_driver.proc_entry  = 0;

	if (tty_register_driver(&serial_driver))
		printk("RCKT550:Couldn't register serial driver\n");

	if (tty_register_driver(&callout_driver))
		printk("RCKT550:Couldn't register callout driver\n");
	
	tprint ( "After registration, MAJOR %d, callout is %d\n", serial_driver.major, callout_driver.major );
	
	if ( DetectComtrolPciBoards() )                     /* returns TRUE if Soft-capable boards are found */
	{
	   RS422_driver             = serial_driver;
	   RS422_driver.name        = "Rckt550-422";
	   RS422_driver.read_proc 	 = 0;
	   RS422_driver.proc_entry  = 0;
		
	   RS422_driver.major = register_chrdev(0,RS422_driver.name,(struct file_operations *)-1);
		
		unregister_chrdev(RS422_driver.major,RS422_driver.name );
	   
		if ( tty_register_driver(&RS422_driver) )
		   printk("RCKT550:Couldn't register RS422 driver\n");
		


	   RS485_driver             = serial_driver;
	   RS485_driver.name        = "Rckt550-485";
	   RS485_driver.read_proc 	 = 0;
   		RS485_driver.proc_entry  = 0;
	   
		RS485_driver.major = register_chrdev(0,RS485_driver.name,(struct file_operations *)-1);
		
		unregister_chrdev(RS485_driver.major,RS485_driver.name );
		
	   if ( tty_register_driver(&RS485_driver) )
		   printk("RCKT550:Couldn't register RS485 driver\n");
	}
	
	for (i = 0, state = rckt550_table; i < NR_PORTS; i++,state++)
	{
		state->magic 				= SSTATE_MAGIC;
		state->line 				= i;
		state->type 				= PORT_UNKNOWN;
		state->custom_divisor 	= 0;
		state->close_delay 		= 5*HZ/10;
		state->closing_wait 		= 30*HZ;
		state->callout_termios 	= callout_driver.init_termios;
		state->normal_termios 	= serial_driver.init_termios;
		state->icount.cts 		= state->icount.dsr =
		state->icount.rng 		= state->icount.dcd = 0;
		state->icount.rx 			= state->icount.tx = 0;
		state->icount.frame 		= state->icount.parity = 0;
		state->icount.overrun 	= state->icount.brk = 0;
		state->irq 					= irq_cannonicalize(state->irq);
		if (check_region(state->port,8))
			continue;
		if (state->flags & ASYNC_BOOT_AUTOCONF)
			autoconfig(state);
	}
	/*
	 * Detect the IRQ only once every port is initialised,
	 * because some 16450 do not reset to 0 the MCR register.
	 */
	for (i = 0, state = rckt550_table; i < NR_PORTS; i++,state++) {
		if (state->type == PORT_UNKNOWN)
			continue;
		if (   (state->flags & ASYNC_BOOT_AUTOCONF)
		    && (state->flags & ASYNC_AUTO_IRQ)
		    && (state->port != 0))
			state->irq = detect_uart_irq(state);
#if EGO_FIX_1
		tprint(KERN_INFO "ttyT%02d%s at 0x%04lx (irq = %d)\n",
		       state->line,
		       (state->flags & ASYNC_FOURPORT) ? " FourPort" : "",
		       state->port, state->irq);
#else
		tprint(KERN_INFO "ttyT%02d%s at 0x%04x (irq = %d)\n",
		       state->line,
		       (state->flags & ASYNC_FOURPORT) ? " FourPort" : "",
		       state->port, state->irq);
#endif
#if RCKT_DEBUG_MORE
	printk (KERN_INFO "EGO:X struct at %X has type=%d\n", (int)state, state->type);
#endif
	}
	return 0;
}

/*
 * register_serial and unregister_serial allows for rckt550 ports to be
 * configured at run-time, to support PCMCIA modems.
 */
int register_serial(struct serial_struct *req)
{
	int i;
	unsigned long flags;
	struct serial_state *state;

	save_flags(flags);
	cli();
	for (i = 0; i < NR_PORTS; i++) {
		if (rckt550_table[i].port == req->port)
			break;
	}
	if (i == NR_PORTS) {
		for (i = 0; i < NR_PORTS; i++)
			if ((rckt550_table[i].type == PORT_UNKNOWN) &&
			    (rckt550_table[i].count == 0))
				break;
	}
	if (i == NR_PORTS) {
		restore_flags(flags);
		return -1;
	}
	state = &rckt550_table[i];
	if (rckt550_table[i].count) {
		restore_flags(flags);
		printk("Couldn't configure rckt550 #%d (port=%d,irq=%d): "
		       "device already open\n", i, req->port, req->irq);
		return -1;
	}
	state->irq = req->irq;
	state->port = req->port;
	state->flags = req->flags;

	autoconfig(state);
	if (state->type == PORT_UNKNOWN) {
		restore_flags(flags);
		printk("register_serial(): autoconfig failed\n");
		return -1;
	}
	restore_flags(flags);

	if ((state->flags & ASYNC_AUTO_IRQ) && (state->port != 0))
		state->irq = detect_uart_irq(state);

#if EGO_FIX_1
	printk(KERN_INFO "tty%02d at 0x%04lx (irq = %d) is a %s\n",
	       state->line, state->port, state->irq, uart_config[state->type].name);
#else
	printk(KERN_INFO "tty%02d at 0x%04x (irq = %d) is a %s\n",
	       state->line, state->port, state->irq, uart_config[state->type].name);
#endif
	return state->line;
}

void unregister_serial(int line)
{
	unsigned long flags;
	struct serial_state *state = &rckt550_table[line];

	save_flags(flags);
	cli();
	if (state->info && state->info->tty)
		tty_hangup(state->info->tty);
	state->type = PORT_UNKNOWN;
	printk(KERN_INFO "tty%02d unloaded\n", state->line);
	restore_flags(flags);
}

#ifdef MODULE
int init_module(void)
{
	return rckt550_init();
}

void cleanup_module(void)
{
	unsigned long flags;
	int e1, e2;
	int i;
	struct async_struct *info;

	/* printk("Unloading %s: version %s\n", serial_name, serial_version); */
	save_flags(flags);
	cli();
        remove_bh(SERIAL_BH);
	if ( serial_driver.major )
	   if ((e1 = tty_unregister_driver(&serial_driver)))
		printk("SERIAL: failed to unregister rckt550 driver (%d)\n", e1);
	if ( callout_driver.major )
	   if ((e2 = tty_unregister_driver(&callout_driver)))
		printk("SERIAL: failed to unregister callout driver (%d)\n", e2);
	if ( RS422_driver.major )
	   if ((e2 = tty_unregister_driver(&RS422_driver)))
		printk("SERIAL: failed to unregister RS422 driver (%d)\n", e2);
	if ( RS485_driver.major )
	   if ((e2 = tty_unregister_driver(&RS485_driver)))
		printk("SERIAL: failed to unregister RS485 driver (%d)\n", e2);
	
	unregister_chrdev(serial_driver.major,"Rckt550d");
	unregister_chrdev(callout_driver.major,"Rckt550m");
	unregister_chrdev(RS422_driver.major,"Rckt550-422");
	unregister_chrdev(RS485_driver.major,"Rckt550-485");
	
	restore_flags(flags);

	for (i = 0; i < NR_PORTS; i++) {
		if (rckt550_table[i].type != PORT_UNKNOWN)
			release_region(rckt550_table[i].port, 8);
		info = rckt550_table[i].info;
		if (info) {
			rckt550_table[i].info = NULL;
#if EGO_FIX_1
			kfree(info);
#else
			kfree_s(info, sizeof(struct async_struct));
#endif
		}
	}
	if (tmp_buf) {
		free_page((unsigned long) tmp_buf);
		tmp_buf = NULL;
	}
}
#endif /* MODULE */


/*
 * ------------------------------------------------------------
 * Serial console driver
 * ------------------------------------------------------------
 */
#ifdef CONFIG_SERIAL_CONSOLE

#define BOTH_EMPTY (UART_LSR_TEMT | UART_LSR_THRE)

/*
 *	Wait for transmitter & holding register to empty
 */
static inline void wait_for_xmitr(struct serial_state *ser)
{
	int lsr;
	unsigned int tmout = 1000000;

	do {
		lsr = inb(ser->port + UART_LSR);
		if (--tmout == 0) break;
	} while ((lsr & BOTH_EMPTY) != BOTH_EMPTY);
}

/*
 *	Print a string to the rckt550 port trying not to disturb
 *	any possible real use of the port...
 */
static void serial_console_write(struct console *co, const char *s,
				unsigned count)
{
	struct serial_state *ser;
	int ier;
	unsigned i;

	ser = rckt550_table + co->index;
	/*
	 *	First save the IER then disable the interrupts
	 */
	ier = inb(ser->port + UART_IER);
	outb(0x00, ser->port + UART_IER);

	/*
	 *	Now, do each character
	 */
	for (i = 0; i < count; i++, s++) {
		wait_for_xmitr(ser);

		/*
		 *	Send the character out.
		 *	If a LF, also do CR...
		 */
		outb(*s, ser->port + UART_TX);
		if (*s == 10) {
			wait_for_xmitr(ser);
			outb(13, ser->port + UART_TX);
		}
	}

	/*
	 *	Finally, Wait for transmitter & holding register to empty
	 * 	and restore the IER
	 */
	wait_for_xmitr(ser);
	outb(ier, ser->port + UART_IER);
}

/*
 *	Receive character from the rckt550 port
 */
static int serial_console_wait_key(struct console *co)
{
	struct serial_state *ser;
	int ier;
	int lsr;
	int c;

	ser = rckt550_table + co->index;

	/*
	 *	First save the IER then disable the interrupts so
	 *	that the real driver for the port does not get the
	 *	character.
	 */
	ier = inb(ser->port + UART_IER);
	outb(0x00, ser->port + UART_IER);

	do {
		lsr = inb(ser->port + UART_LSR);
	} while (!(lsr & UART_LSR_DR));
	c = inb(ser->port + UART_RX);

	/*
	 *	Restore the interrupts
	 */
	outb(ier, ser->port + UART_IER);

	return c;
}

static kdev_t serial_console_device(struct console *c)
{
	return MKDEV(TTY_MAJOR, 64 + c->index);
}

/*
 *	Setup initial baud/bits/parity. We do two things here:
 *	- construct a cflag setting for the first rckt550_open()
 *	- initialize the rckt550 port
 *	Return non-zero if we didn't find a rckt550 port.
 */
#if EGO_FIX_1
__init static int serial_console_setup(struct console *co, char *options)
#else
__initfunc(static int serial_console_setup(struct console *co, char *options))
#endif
{
	struct serial_state *ser;
	unsigned cval;
	int	baud = 9600;
	int	bits = 8;
	int	parity = 'n';
	int	cflag = CREAD | HUPCL | CLOCAL;
	int	quot = 0;
	char	*s;

	if (options) {
		baud = simple_strtoul(options, NULL, 10);
		s = options;
		while(*s >= '0' && *s <= '9')
			s++;
		if (*s) parity = *s++;
		if (*s) bits   = *s - '0';
	}

	/*
	 *	Now construct a cflag setting.
	 */
	switch(baud) {
		case 1200:
			cflag |= B1200;
			break;
		case 2400:
			cflag |= B2400;
			break;
		case 4800:
			cflag |= B4800;
			break;
		case 19200:
			cflag |= B19200;
			break;
		case 38400:
			cflag |= B38400;
			break;
		case 57600:
			cflag |= B57600;
			break;
		case 115200:
			cflag |= B115200;
			break;
		case 9600:
		default:
			cflag |= B9600;
			break;
	}
	switch(bits) {
		case 7:
			cflag |= CS7;
			break;
		default:
		case 8:
			cflag |= CS8;
			break;
	}
	switch(parity) {
		case 'o': case 'O':
			cflag |= PARODD;
			break;
		case 'e': case 'E':
			cflag |= PARENB;
			break;
	}
	co->cflag = cflag;

	/*
	 *	Divisor, bytesize and parity
	 */
	ser = rckt550_table + co->index;
	quot = ser->baud_base / baud;
	cval = cflag & (CSIZE | CSTOPB);
#if defined(__powerpc__) || defined(__alpha__)
	cval >>= 8;
#else /* !__powerpc__ && !__alpha__ */
	cval >>= 4;
#endif /* !__powerpc__ && !__alpha__ */
	if (cflag & PARENB)
		cval |= UART_LCR_PARITY;
	if (!(cflag & PARODD))
		cval |= UART_LCR_EPAR;

	/*
	 *	Disable UART interrupts, set DTR and RTS high
	 *	and set speed.
	 */
	outb(cval | UART_LCR_DLAB, ser->port + UART_LCR);	/* set DLAB */
	outb(quot & 0xff, ser->port + UART_DLL);	/* LS of divisor */
	outb(quot >> 8, ser->port + UART_DLM);		/* MS of divisor */
	outb(cval, ser->port + UART_LCR);		/* reset DLAB */
	outb(0, ser->port + UART_IER);
	outb(UART_MCR_DTR | UART_MCR_RTS, ser->port + UART_MCR);

	/*
	 *	If we read 0xff from the LSR, there is no UART here.
	 */
	if (inb(ser->port + UART_LSR) == 0xff)
		return -1;
	return 0;
}

static struct console sercons = {
	"ttyT",
	serial_console_write,
	NULL,
	serial_console_device,
	serial_console_wait_key,
	NULL,
	serial_console_setup,
	CON_PRINTBUFFER,
	-1,
	0,
	NULL
};

/*
 *	Register console.
 */
#if EGO_FIX_1
__init long serial_console_init(long kmem_start, long kmem_end)
#else
__initfunc (long serial_console_init(long kmem_start, long kmem_end))
#endif
{
	register_console(&sercons);
	return kmem_start;
}
#endif


